from typing import Optional

from chemrecon.schema.enums import SourceDatabase
from chemrecon.schema.db_object import Relation, col_src, InverseRelation
from chemrecon.schema.entry_types.compound import Compound


# Is A / Has Instance
# ----------------------------------------------------------------------------------------------------------------------
class CompoundIsA(Relation[Compound, Compound]):
    """ Hierarchical relation which indicates that a compound is a member of a class of compounds.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Is a'
    _table_name = 'CompoundIsA'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'is a'

class CompoundHasInstance(InverseRelation[Compound, Compound]):
    """ Hierarchical relation which indicates that a compound is a member of a class of compounds.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Inverse
    inverse_main_relation = CompoundIsA

    # Database
    entrytype_name = 'Has Instance'
    _table_name = 'CompoundHasInstance'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'has instance'

# Set Inverse
CompoundIsA.has_inverse = CompoundHasInstance

# New / Old id
# ----------------------------------------------------------------------------------------------------------------------
class CompoundHasNewID(Relation[Compound, Compound]):
    """ Indicates correspondence between identifiers in different database versions. Can be used to resolve deprecated
        identifiers.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Has New ID'
    _table_name = 'CompoundHasNewID'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'new id'

class CompoundHasOldID(InverseRelation[Compound, Compound]):
    """ Indicates correspondence between identifiers in different database versions. Can be used to resolve deprecated
        identifiers.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Inverse
    inverse_main_relation = CompoundHasNewID

    # Database
    entrytype_name = 'Has Old ID'
    _table_name = 'CompoundHasOldID'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'old id'

# Set Inverse
CompoundHasNewID.has_inverse = CompoundHasOldID


# Has Part / Is Part of
# ----------------------------------------------------------------------------------------------------------------------
class CompoundHasPart(Relation[Compound, Compound]):
    """ Represents a relationship where one compound is a part of another.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Has Part'
    _table_name = 'CompoundHasPart'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'has part'

class CompoundIsPartOf(InverseRelation[Compound, Compound]):
    """ Represents a relationship where one compound is a part of another.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Inverse
    inverse_main_relation = CompoundHasNewID

    # Database
    entrytype_name = 'Is Part of'
    _table_name = 'CompoundIsPartOf'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'is part of'

# Set Inverse
CompoundHasPart.has_inverse = CompoundIsPartOf


# Conjugate Acid Base
# ----------------------------------------------------------------------------------------------------------------------
class CompoundHasConjugateAcid(Relation[Compound, Compound]):
    """ Represents a relationship indicating that a compound has a conjugate acid.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Has Conjugate Acid'
    _table_name = 'CompoundHasConjugateAcid'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'conj. acid'

class CompoundHasConjugateBase(InverseRelation[Compound, Compound]):
    """ Represents a relationship indicating that a compound has a conjugate base.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Inverse
    inverse_main_relation = CompoundHasNewID

    # Database
    entrytype_name = 'Has Conjugate Base'
    _table_name = 'CompoundHasConjugateBase'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'conj. base'

# Set Inverse
CompoundHasConjugateAcid.has_inverse = CompoundHasConjugateBase

# Tautomer
# ----------------------------------------------------------------------------------------------------------------------
class CompoundHasTautomer(Relation[Compound, Compound]):
    """ Represents that two compounds are considered tautomers of each other.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Has Tautomer'
    _table_name = 'CompoundHasTautomer'
    symmetric = True
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'tautomer'

# Stereoisomer
# ----------------------------------------------------------------------------------------------------------------------
class CompoundHasStereoIsomer(Relation[Compound, Compound]):
    """ Represents that two compounds are considered stereoisomers of each other.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Has Stereoisomer'
    _table_name = 'CompoundHasStereoisomer'
    symmetric = True
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'stereoisomer'

# Isotopologue
# ----------------------------------------------------------------------------------------------------------------------
class CompoundHasIsotopologue(Relation[Compound, Compound]):
    """ Represents that two compounds are considered isotopologues of each other.
        (identical in elements, but different isotopic composition).
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Has Isotopologue'
    _table_name = 'CompoundHasIsotopologue'
    symmetric = True
    source_entrytype = Compound
    target_entrytype = Compound
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'isotopologue'




from __future__ import annotations

from typing import Optional, ClassVar, Generator

import rustworkx as rx

from chemrecon.entrygraph.draw import defaultdrawer
from chemrecon.entrygraph.explore_procedure import ExploreProcedure
from chemrecon.schema import Entry, Relation

from chemrecon import connection as connection

type VertexIndex = int
type EdgeIndex = int
type ReconID = int


# Vertices and Edges
# --------------------------------------------------------------------------------------------------------------
# Vertex and Edge objects are used as the payload for the RustWorkX graph.

class VertexBase:
    pass


class Vertex[E: Entry](VertexBase):
    """ A database entry in a graph context.
        None is used as the type parameter for artificial vertices (those not corresponding to a DB object).
        A vertex is initial if its generation is 0.
    """
    entry: E  #: The entry represented by this vertex
    recon_id: ReconID  #:
    vertex_index: VertexIndex  #: Index in the graph representation.
    generation: int  #: Distance from initial entries.

    def __init__(self, entry: E, generation: int):
        assert entry.recon_id is not None
        self.entry = entry
        self.recon_id = entry.recon_id
        self.generation = generation


class EdgeBase:
    pass


class Edge[R: Relation | None](EdgeBase):
    """ A database relation in a graph context.
        None is used as the type parameter for artificial edges (those not corresponding to a DB object).
    """
    relation: R  #:
    recon_id_1: ReconID  #:
    recon_id_2: ReconID  #:
    edge_index: EdgeIndex  #: Index in the graph representation

    def __init__(self, relation: R):
        self.relation = relation
        self.recon_id_1 = relation.recon_id_1
        self.recon_id_2 = relation.recon_id_2


# Special case graph vertices and edges which do not represent database entries or relations
class SourceVertexArtificial(VertexBase):
    """ Artificial vertex used as a source"""

    def __init__(self):
        pass


class SourceEdgeArtificial(EdgeBase):
    """ Artificial edge from the SourceVertex to initial vertices in the graph. """

    def __init__(self):
        pass


class ReturnEdgeArtificial(EdgeBase):
    """ Temporary edge added from sink vertices back to the SourceVertex for pagerank purposes. """

    def __init__(self):
        pass


# Generic EntryGraph
# --------------------------------------------------------------------------------------------------------------
class EntryGraph:
    """
    Represents a directed graph structure of entries and relations.

    :ivar g: RustWorkX directed graph object used to store vertices and edges.
    :type g: rx.PyDiGraph
    :ivar index_entry: Maps Entry objects to corresponding vertex indices.
    :type index_entry: dict[Entry, VertexIndex]
    :ivar index_reconid: Maps (Entry type, ReconID) tuples to vertex indices.
    :type index_reconid: dict[tuple[type[Entry], ReconID], VertexIndex]
    :ivar index_relation: Maps Relation objects to corresponding edge indices.
    :type index_relation: dict[Relation, EdgeIndex]
    :ivar initial_entries: Set of Entry objects that serve as the initial vertices in the graph.
    :type initial_entries: set[Entry]
    :ivar initial_vertices: Set of vertex indices corresponding to the initial entries.
    :type initial_vertices: set[VertexIndex]
    :ivar parent_entrygraph: Reference to the parent EntryGraph if this graph
        is a subgraph view.
    :type parent_entrygraph: Optional[EntryGraph]
    """
    g: rx.PyDiGraph

    # Indices for looking up vertices / edges based on the underlying entries
    index_entry: dict[Entry, VertexIndex]
    index_reconid: dict[tuple[type[Entry], ReconID], VertexIndex]
    index_relation: dict[Relation, EdgeIndex]

    # Misc
    initial_entries: set[Entry]
    initial_vertices: set[VertexIndex]

    # Parentage
    parent_entrygraph: Optional[EntryGraph]  # If created as a subgraph view, gives the parent graph

    # List of finished ExplorationProcedure objects to call when exploring
    explore_procedures: ClassVar[list[ExploreProcedure]]
    post_explore_procedures: ClassVar[list[ExploreProcedure]]
    transitive_subprocedures: ClassVar[dict[type[Relation], tuple[ExploreProcedure, ExploreProcedure]]]

    # TODO scoring as abstract class variables


    def __init__(
        self,  # TODO specification refactor
        initial_entries: set[Entry]
    ):
        """ Initialize an entrygraph with a set of initial vertices.
        """
        # Consistency check
        # TODO init vertices in allowed types

        # Create
        self.index_entry = dict()
        self.index_reconid = dict()
        self.index_relation = dict()

        # Create graph
        self.g = rx.PyDiGraph(multigraph = True)

        # Add initial vertices
        # The initial vertices may not have recon-ids attached, so we add those by DB lookup
        for entry in initial_entries:
            if entry.recon_id is None:
                res = connection.handler.get_entry_by_index(entry)
                if res is None:
                    continue
                entry.recon_id = res.recon_id

        # If no initial vertices have recon ids, abort
        if all(e.recon_id is None for e in initial_entries):
            raise ValueError('No initial entries have recon ids.')

        self.initial_entries = initial_entries
        self.initial_vertices = {
            self.add_vertex(e, generation = 0)
            for e in initial_entries
        }

    # Getters
    # ----------------------------------------------------------------------------------------------------------
    def get_vertex_index_by_recon_id(self, entry_type: type[Entry], recon_id: ReconID) -> Optional[VertexIndex]:
        """
        Retrieve the vertex index associated with the given entry type and recon ID.
        If no matching vertex index is found, the method returns None.

        :param entry_type: The class type of the entry being queried.
        :type entry_type: type[Entry]
        :param recon_id: The recon ID that uniquely identifies the entry.
        :type recon_id: ReconID
        :return: The vertex index corresponding to the given entry type and recon ID, or None
                 if not found.
        :rtype: Optional[VertexIndex]
        """
        return self.index_reconid.get((entry_type, recon_id), None)

    def get_vertex_by_entry(self, entry: Entry) -> Optional[VertexIndex]:
        """
        Retrieves the vertex index associated with a given entry and returns None if it does not exist.

        :param entry: The entry object to search for in the index map.
        :type entry: Entry
        :return: The vertex index associated with the given entry, or None if the entry
                 does not exist in the mapping.
        :rtype: Optional[VertexIndex]
        """
        return self.index_entry.get(entry, None)

    def get_vertex_by_vertex_index(self, vertex_index: VertexIndex) -> Optional[Vertex]:
        """
        Retrieve the vertex corresponding to the given vertex index, returns None if it does not exist.
        :param vertex_index: Index of the vertex to retrieve.
        :type vertex_index: VertexIndex
        :return: The vertex associated with the given index or None if it does not exist.
        :rtype: Optional[Vertex]
        """
        return self.g.get_node_data(vertex_index)

    def get_out_edges_of_vertex(self, vertex_index: VertexIndex) -> list[tuple[Edge, Vertex]]:
        """
        Retrieve all outgoing edges of a vertex along with the corresponding target
        vertices.

        This method provides a list of tuples where each tuple contains the edge data
        and the target vertex data for the edges originating from the specified
        vertex. This function is useful for analysing the outgoing connections
        and relationships in a graph data structure.

        :param vertex_index: Index of the vertex whose outgoing edges are to be queried.
        :type vertex_index: VertexIndex
        :return: A list of tuples with each tuple containing edge data and the target
                 vertex data.
        :rtype: list[tuple[Edge, Vertex]]
        """
        return [
            (edge_data, self.g.get_node_data(target_index))
            for _, target_index, edge_data in self.g.out_edges(vertex_index)
        ]

    # Iterators
    def vertices(self) -> Generator[Vertex, None, None]:
        """ Generator over vertices (entries). """
        yield from self.g.nodes()

    def edges(self) -> Generator[Edge, None, None]:
        """ Generator over edges (relations). """
        yield from self.g.edges()

    # Adders
    # ----------------------------------------------------------------------------------------------------------
    def add_vertex(self, entry: Entry, generation: int = -1) -> VertexIndex:
        """ Add a new entry to the graph. If it already exists, return the index.
        """
        if entry.recon_id is None:
            raise ValueError('Cannot construct vertex from entry without reconid.')
        lookup = self.get_vertex_index_by_recon_id(type(entry), entry.recon_id)
        if lookup is not None:
            return lookup

        # Add the vertex
        vertex = Vertex(entry, generation = generation)
        v_index: VertexIndex = self.g.add_node(vertex)
        self.index_reconid[(type(entry), entry.recon_id)] = v_index
        self.index_entry[entry] = v_index
        vertex.vertex_index = v_index
        return v_index

    def add_edge(
        self,
        source_v_index: VertexIndex,
        target_v_index: VertexIndex,
        relation: Relation
    ) -> EdgeIndex:
        """ Add a relation between the specified source and target vertices
        """
        # Try returning existing edge
        try:
            return self.index_relation[relation]
        except KeyError:
            # Does not already exist
            edge = Edge(relation)
            edge_index: EdgeIndex = self.g.add_edge(source_v_index, target_v_index, edge)
            edge.edge_index = edge_index
            self.index_relation[relation] = edge_index
            return edge_index

    def add_vertex_from[T1: Entry, T2: Entry](
        self,
        from_index: VertexIndex,
        relation: Relation[T1, T2],
        entry: T2,
        generation: int
    ) -> VertexIndex:
        """ Adds a new entry vertex along with a relation. Returns vertex_index of the newly creted vertex.
            If the vertex already exists, return add the edge and return the vertex index.
        """
        target_v_index = self.add_vertex(entry, generation = generation)
        self.add_edge(from_index, target_v_index, relation)
        return target_v_index

    # Draw
    # ----------------------------------------------------------------------------------------------------------
    def draw(
        self,
        filename: str = None,
        filetype: str = 'jpg',
        scores: Optional[dict[Entry, float]] = None
    ):
        """ Draw using default settings, returning an image, or write to the disk if the filename is specified.
            For more settings, create an EntryGraphDrawer instance.
        """
        return defaultdrawer.draw(self, filename = filename, filetype = filetype, scores = scores)

    def show(self, scores: Optional[dict[Entry, float]] = None):
        """ Same as draw, but displays the image in a window instead of saving to disk.
        """
        return self.draw(scores = scores).show()



""" Method for creating an explored Entry Graph based on the database.
"""
from typing import Optional

import psycopg as pg

import chemrecon.connection as connection

from chemrecon.schema import Entry, Relation, ComposedRelation, ProceduralRelation, ProceduralGeneratorError

from chemrecon.entrygraph.entrygraph import EntryGraph, ReconID, VertexIndex
from chemrecon.entrygraph.explore_procedure import (
    ExploreProcedure, ExploreProcedureT1, ExploreProcedureT2, ExploreProcedureSym
)
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol


def explore(
    entrygraph: EntryGraph,
    protocol: ExplorationProtocol,
    steps: int = 4
):
    """ Expand the given entry graph by traversing the database network using the specified protocol for a given
        number of steps.
    """
    Explorer(entrygraph, protocol, steps).explore()


class Explorer:
    entrygraph: EntryGraph
    generations: Optional[int] = None
    protocol: ExplorationProtocol

    def __init__(
        self,
        entrygraph: EntryGraph,
        protocol: ExplorationProtocol,
        generations: int = None
    ):
        self.entrygraph = entrygraph
        self.generations = generations
        self.protocol = protocol

        # Generate cursors for each procedure
        for proc in {
            *protocol.explore_procedures,
            *protocol.post_explore_procedures,
            *[p for p, _ in protocol.transitive_subprocedures.values()],
            *[p for _, p in protocol.transitive_subprocedures.values()]
        }:
            proc: ExploreProcedure
            proc.cursor = pg.Cursor(
                connection = connection.handler.conn,
                row_factory = connection.handler.make_relation_entry_view_row_factory(proc.relation_type)
            )

    # Main exploration alg.
    # --------------------------------------------------------------------------------------------------------------
    def explore(self) -> None:
        """ Explores the graph based on the database connection.
            After execution, the graph will be fully populated.
        """
        import chemrecon.core.populate_query_handler

        # Sets of already explored vertices (updated at end of every iteration)
        vs_new: set[VertexIndex] = set()
        vs_explored: set[VertexIndex] = set()

        # Add initial vertices
        for v_idx in self.entrygraph.initial_vertices:
            vs_new.add(v_idx)

        # Iteratively explore
        i: int = 0  # Generation
        while True:
            # Main iteration
            i += 1

            # Compute new set of vertices to explore
            to_explore: dict[type[Entry], set[tuple[VertexIndex, ReconID]]] = {
                etype: set() for etype in self.protocol.entry_types
            }
            for v_idx in vs_new:
                v = self.entrygraph.get_vertex_by_vertex_index(v_idx)
                if v is None:
                    assert False, 'unreachable'
                to_explore[type(v.entry)].add((v_idx, v.recon_id))

            vs_explored.update(vs_new)

            # For each type of relation, call the explore procedure, and sort out entries according to the filter
            for explr_proc in self.protocol.explore_procedures:

                # If past generation limit, explore only relations which explicitly ignore this
                if i > self.generations:
                    if not explr_proc.relation_type.ignore_generation_limit:
                        continue

                # Run procedure, producing relations and entry endpoints
                input_entries = list(to_explore[explr_proc.takes_entrytype])
                if len(input_entries) > 0:
                    try:
                        res = self._run_procedure(
                            explr_proc,
                            [self.entrygraph.g.get_node_data(v).entry for v, _ in input_entries]
                        )
                        if res is None:
                            # Should not happen (?)
                            continue
                    except ProceduralGeneratorError:
                        # Could not generate, so do not add
                        continue

                    # res: list of (t_takes, relation[t_takes, t_gives], t_gives)
                    for (takes_vertex_index, takes_recon_id), subresult in zip(input_entries, res):
                        # TODO error here, sym relations continually gives the e_takes instead of the e_gives
                        for rel, e_gives in subresult:

                            # Filter
                            if explr_proc.relation_filter is not None:
                                if not explr_proc.relation_filter(rel):
                                    continue
                            if explr_proc.entry_filter is not None:
                                if not explr_proc.entry_filter(e_gives):
                                    continue

                            # Add relation to entrygraph
                            v_new_idx = self.entrygraph.add_vertex_from(
                                from_index = takes_vertex_index,
                                relation = rel,
                                entry = e_gives,
                                generation = i
                            )
                            vs_new.add(v_new_idx)  # Add new vertices to explore

            # End of iteration, update lists
            vs_new = vs_new - vs_explored
            if len(vs_new) == 0:
                break

        # Explore terminal relation (does not lead to any new vertices to add)
        # TODO should not add new vertices, only relations between vertices already in the graph
        # TODO !important
        for explr_proc in self.protocol.post_explore_procedures:
            input_entries: list[tuple[VertexIndex, ReconID]] = [
                (v.vertex_index, v.recon_id)
                for v in self.entrygraph.g.nodes()
                if type(v.entry) is explr_proc.takes_entrytype
            ]
            if len(input_entries) > 0:
                res = self._run_procedure(
                    explr_proc,
                    [self.entrygraph.g.get_node_data(v).entry for v, _ in input_entries]
                )
                for (takes_vertex_index, takes_recon_id), subresult in zip(input_entries, res):
                    for rel, e_gives in subresult:
                        # Add relation to entrygraph only if exists
                        lookup = self.entrygraph.get_vertex_by_entry(e_gives)
                        if lookup is not None:
                            self.entrygraph.add_edge(
                                source_v_index = takes_vertex_index,
                                target_v_index = lookup,
                                relation = rel
                            )

                        # TODO add setting to create new entries from terminal relations?
                        # Old, added new entries unnecessarily
                        # v_new_idx = self.entrygraph.add_vertex_from(
                        #     from_index = takes_vertex_index,
                        #     relation = rel,
                        #     entry = e_gives,
                        #     generation = -1
                        # )

        # Commit changes made if handler can populate the backend
        if isinstance(connection.handler, chemrecon.core.populate_query_handler.PopulateQueryHandler):
            connection.handler.conn.commit()

        # Done, finalise
        pass


    # Dispatchers for individual procedures
    # ---------------------------------------------------------------------------------------------------------------
    def _run_procedure[T_takes: Entry, T_gives: Entry](
        self,
        proc: ExploreProcedure[T_takes, T_gives],
        take_entries: list[T_takes]
    ) -> list[list[tuple[Relation[T_takes, T_gives], T_gives]]]:
        if len(take_entries) == 0:
            return []
        # Dispatch depending on whether procedural or database-backed relation
        if issubclass(proc.relation_type, ComposedRelation):
            return self._run_procedure_transitive(proc, take_entries)
        elif issubclass(proc.relation_type, ProceduralRelation):
            return self._run_procedure_procedural(proc, take_entries)
        else:
            return self._run_procedure_database(proc, take_entries)


    def _run_procedure_database[T_takes: Entry, T_gives: Entry](
        self,
        proc: ExploreProcedure[T_takes, T_gives],
        take_entries: list[T_takes],
    ) -> list[list[tuple[Relation[T_takes, T_gives], T_gives]]]:
        # The result of the procedure relies only on information in the database
        proc.cursor.executemany(
            query = proc.q,
            params_seq = [[e.recon_id] * proc.n_params for e in take_entries],
            returning = True
        )

        # Fetch results and return
        result: list[list[tuple[Relation, Entry, Entry]]] = [proc.cursor.fetchall()]
        while proc.cursor.nextset():
            result.append(proc.cursor.fetchall())

        # How to interpret and add depends on T1, T2, or sym relation
        match proc:
            case ExploreProcedureT1():
                # Result is relation and T2
                return [
                    [(r, e2) for r, e1, e2 in subresult]
                    for subresult in result
                ]
            case ExploreProcedureT2():
                # Result is relation and T1
                return [
                    [(r, e1) for r, e1, e2 in subresult]
                    for subresult in result
                ]
            case ExploreProcedureSym():
                # Result entry is T1 or T2, depending on which is different from the input
                return [
                    [
                        (r, e2) if e1.recon_id == takes_recon_id else (r, e1)
                        for r, e1, e2 in subresult
                    ]
                    for takes_recon_id, subresult in zip((
                        e.recon_id for e in take_entries
                    ), result)
                ]
            case _:
                assert False, 'unreachable'


    def _run_procedure_procedural[T_takes: Entry, T_gives: Entry](
        self,
        proc: ExploreProcedure[T_takes, T_gives],
        take_entries: list[T_takes],
    ) -> list[list[tuple[Relation[T_takes, T_gives], T_gives]]]:
        import chemrecon.core.populate_query_handler

        # TODO handle catching the ProceduralGeneratorError()!

        # Run the generate() method of the relation, and possibly update the DB
        # if the handler supports insertion
        assert issubclass(proc.relation_type, ProceduralRelation)

        # First, check that the relation already exists in the DB.
        db_result = self._run_procedure_database(proc, take_entries)

        # If we found results for all queries
        if all(len(subresult) > 0 for subresult in db_result):
            return db_result

        # If not , use generator
        procedural_result: list[list[tuple[Relation[T_takes, T_gives], T_gives]]] = list()

        for i, take_entry in enumerate(take_entries):
            if len(db_result[i]) > 0:
                procedural_result.append(db_result[i])
                continue
            res = proc.relation_type.generate(
                take_entry
            )
            procedural_result.append([
                (rel, e_gives)
                for rel, e_gives in res
            ])

        # Update the database based on the db_result if possible
        if isinstance(connection.handler, chemrecon.core.populate_query_handler.PopulateQueryHandler):
            for take_reconid, subresult in zip((e.recon_id for e in take_entries), procedural_result):
                assert take_reconid is not None
                if len(subresult) == 0:
                    continue

                assigned_ids = connection.handler.add_relations_to_entry_with_reconid(
                    recon_id = take_reconid,
                    entry_table = proc.relation_type.source_entrytype,
                    relations = [
                        (rel, e_gives)
                        for rel, e_gives in subresult
                    ]
                )

                # Assign recon_ids to the created entries
                for assign_recon_id, (new_rel, new_entry) in zip(assigned_ids, subresult):
                    new_entry.recon_id = assign_recon_id

        else:
            # Assign virtual Recon IDs to the generated entries (do not necessarily correspond with db)
            # These will be negative numbers
            for subresult in procedural_result:
                for rel, e in subresult:
                    connection.handler.add_procedural_entry(e)

        # Return result
        return procedural_result


    def _run_procedure_transitive[T_takes: Entry, T_gives: Entry, T_intermediate: Entry](
        self,
        proc: ExploreProcedure[T_takes, T_gives],
        take_entries: list[T_takes],
    ) -> list[list[tuple[Relation[T_takes, T_gives], T_gives]]]:
        import chemrecon.core.populate_query_handler

        assert issubclass(proc.relation_type, ComposedRelation)

        # First, check that the relation already exists in the DB.
        db_result = self._run_procedure_database(proc, take_entries)

        # If we found results for all queries
        if all(len(subresult) > 0 for subresult in db_result):
            return db_result

        # Get rel1 results
        rel_1_proc, rel_2_proc = self.protocol.transitive_subprocedures[proc.relation_type]
        rel_1_result = self._run_procedure(rel_1_proc, take_entries)

        # Filter by rel1 and e_inter
        rel_1_result_filtered: list[list[tuple[Relation[T_takes, T_intermediate], T_intermediate]]] = list()
        for subresult_1 in rel_1_result:
            subresult_1_filtered: list[tuple[Relation[T_takes, T_intermediate], T_intermediate]] = list()
            for r1, e_inter in subresult_1:
                if proc.relation_type.filter_rel_1(r1) and proc.relation_type.filter_intermediate(e_inter):
                    subresult_1_filtered.append((r1, e_inter))
            rel_1_result_filtered.append(subresult_1_filtered)

        # Results of rel1 are passed as take_entries to rel_2
        intermediate_map: dict[T_takes, list[tuple[Relation[T_takes, T_intermediate], T_intermediate]]] = {
            e_take: subresult
            for e_take, subresult in zip(take_entries, rel_1_result_filtered)
        }
        intermediate_entries: set[T_intermediate] = set()
        for rels in intermediate_map.values():
            intermediate_entries.update(e for _, e in rels)
        intermediate_entries_list = [e for e in intermediate_entries if proc.relation_type.filter_intermediate(e)]

        # Get rel2 results
        rel_2_result = self._run_procedure(rel_2_proc, intermediate_entries_list)
        rel_2_result_dict: dict[T_intermediate, list[tuple[Relation[T_intermediate, T_gives], T_gives]]] = {
            e_inter: subresult
            for e_inter, subresult in zip(intermediate_entries_list, rel_2_result)
        }

        # TODO filter rel2 results (not currently used)

        # Make list of entries to return
        result: list[list[tuple[Relation[T_takes, T_gives], T_gives]]] = list()
        for i, take_entry in enumerate(take_entries):
            # Consider routes through all intermediates
            subresult: list[tuple[Relation[T_takes, T_gives], T_gives]] = list()
            for rel_1, e_inter in intermediate_map[take_entry]:
                # Add all from this intermediate
                subresult.extend(
                    (proc.relation_type(rel_1 = rel_1, rel_2 = rel_2, intermediate = e_inter), e_gives)
                    for rel_2, e_gives in rel_2_result_dict[e_inter]
                )
            result.append(subresult)


        if isinstance(connection.handler, chemrecon.core.populate_query_handler.PopulateQueryHandler):
            # Update the database from rel_1 if procedural
            for take_entry, subresult in zip(take_entries, rel_1_result_filtered):
                assert take_entry.recon_id is not None
                assigned_ids = connection.handler.add_relations_to_entry_with_reconid(
                    recon_id = take_entry.recon_id,
                    entry_table = proc.relation_type.source_entrytype,
                    relations = [
                        (rel, e_gives)
                        for rel, e_gives in subresult
                    ]
                )
                for assign_recon_id, (new_rel, new_entry) in zip(assigned_ids, subresult):
                    new_entry.recon_id = assign_recon_id

            # Update the database from rel_2 if procedural
            for intermediate_entry, subresult in zip(intermediate_entries_list, rel_2_result):
                assert intermediate_entry.recon_id is not None
                assigned_ids = connection.handler.add_relations_to_entry_with_reconid(
                    recon_id = intermediate_entry.recon_id,
                    entry_table = proc.relation_type.intermediate_entrytype,
                    relations = [
                        (rel, e_gives)
                        for rel, e_gives in subresult
                    ]
                )
                for assign_recon_id, (new_rel, new_entry) in zip(assigned_ids, subresult):
                    new_entry.recon_id = assign_recon_id

            # Update the database for the transitive relation
            for take_entry, subresult in zip(take_entries, result):
                connection.handler.add_relations_to_entry_with_reconid(
                    recon_id = take_entry.recon_id,
                    entry_table = proc.relation_type.source_entrytype,
                    relations = [
                        (rel, e_gives)
                        for rel, e_gives in subresult
                    ]
                )

        else:
            # Assign virtual reconids to the generated entries
            for subresult in result:
                for rel, e in subresult:
                    connection.handler.add_procedural_entry(e)

        return result



from __future__ import annotations

import networkx as nx

from chem.elements import ChemGraph, chemgraph_from_mol, ematch, nmatch
from chem.elements import mol_from_structurerepresentation
from chemrecon.core.query_handler import QueryHandler
from chemrecon.schema.entry_types.compound import Compound
from chemrecon.schema.entry_types.reaction import Reaction
from chemrecon.schema.relation_types_source.compound_has_structure_representation_relation import \
    CompoundHasStructureRepresentation
from chemrecon.schema.relation_types_source.reaction_involves_compound_relation import ReactionInvolvesCompound
from utils import hungarian as hungarian


# TODO change to use the general reactant_matching with ChemReactions

def match_reactants(
    handler: QueryHandler,
    procedure: Procedure,
    chemgraph: ChemGraph,
    reaction: Reaction,
    side: int  # -1 or 1
) -> dict[frozenset[int], tuple[Compound, int]]:
    """ Given a left or right S_GML of a reaction (already sanitised by removing enzymes), tries to match the
        components of the graph to a list of entries with known structures using graph similarity. Returns a
        mapping, from each reactant (compound, multiplicity index), to the set of nodes with which they
        were identified.
    """

    # TODO eventually replace the structure getting alg. with entry graph-based approach

    # Get participant compounds
    reaction_all_compounds: list[tuple[Compound, int]] = list()  # With stoich
    compound_result_dict = handler.get_relations_with_entries_by_recon_ids_of_t1(
        entry_type = Reaction,
        recon_ids = [reaction.recon_id],
        relation_type = ReactionInvolvesCompound,
    )[0]
    for compound_entry, relations in compound_result_dict.items():
        relations: list[ReactionInvolvesCompound]
        for relation in relations:
            reaction_all_compounds.append(
                (compound_entry, relation.n)
            )

    # Get existing structures for the attached compounds (converted to chemgraph objects?)
    compound_graphs: dict[Compound, ChemGraph] = dict()
    for compound_entry, stoich in reaction_all_compounds:
        # Only take compounds on the correct side
        if stoich * side < 0:
            continue

        # Get structures
        struct_repr_relations = handler.get_relations_with_entries_by_recon_ids_of_t1(
            entry_type = Compound,
            recon_ids = [compound_entry.recon_id],
            relation_type = CompoundHasStructureRepresentation
        )
        struct_reprs = struct_repr_relations[0].keys()

        # For now, select an arbitrary structure
        try:
            struct_repr = list(struct_reprs)[0]
        except IndexError:
            # No structures found
            continue

        # Convert the selected structure to a S_GML ChemGraph
        compound_graphs[compound_entry] = chemgraph_from_mol(mol_from_structurerepresentation(struct_repr))

    # Perform the matching algorithm on match_structures against components of the given chemgraph
    chemgraph.remove_hydrogens()
    component_subsets: list[frozenset[int]] = chemgraph.get_component_subsets()
    component_subgraphs: list[nx.Graph] = chemgraph.get_components_subgraphs()

    # Log an error if component numbers don't match
    if len(component_subgraphs) != len(compound_graphs):
        # TODO Try to fix this by balancing cofactors etc
        # TODO May be extra hydrons, water etc
        # TODO  - entry #6 has an extra water on both sides for instance
        procedure.log_notice(f'Could not get structures for all components.'
                             f' Reaction {reaction} (side {side}).'
                             f' Number of S_GML graph components ({len(component_subgraphs)})'
                             f' does not match number of reactants ({len(compound_graphs)}).')

    # Pairwise similarity (maps from each compound to the index of the component and the similarity)
    graph_distance_pairwise: dict[Compound, list[tuple[int, float]]] = dict()

    # Compute the similarity
    for compound_entry, compound_chemgraph in compound_graphs.items():
        component_distances: list[tuple[int, float]] = list()
        for component_index, component in enumerate(component_subgraphs):
            # nx.optimize_graph_edit_distance is a generator which generates successively more precise values.
            dist = next(nx.optimize_graph_edit_distance(
                compound_chemgraph.graph,
                component,
                node_match = nmatch,
                edge_match = ematch
            ))
            component_distances.append((component_index, dist))
        graph_distance_pairwise[compound_entry] = component_distances

    # We make duplicate instances of each compound to match if stoich != 1
    compound_instances: dict[tuple[Compound, int], ChemGraph] = dict()
    for compound, stoich in reaction_all_compounds:
        if stoich * side < 0:
            continue
        for i in range(abs(stoich)):
            try:
                compound_instances[(compound, i)] = compound_graphs[compound]
            except KeyError as e:
                pass

    # Run the hungarian algorithm to get the best match with the instanced compounds
    edges: dict[tuple[tuple[Compound, int], int], float] = dict()
    for (compound, compound_instance_index), _ in compound_instances.items():
        for component_index, distance in graph_distance_pairwise[compound]:
            edges[((compound, compound_instance_index), component_index)] = distance
    matching = hungarian.max_weight_matching(
        edges = edges,
        min_weight = True
    )
    # TODO extra logic to iterate the edit distance algorithm if the matching is 'close'?

    matching_dict: dict[set[int], tuple[Compound, int]] = dict()
    for (compound, compound_index), component_index in matching.items():
        matching_dict[component_subsets[component_index]] = (compound, compound_index)

    return matching_dict



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, ClassVar

import psycopg as pg
from psycopg import sql as sql

from chemrecon.entrygraph.filter import EntryFilter, RelationFilter
from chemrecon.schema.db_object import Entry, Relation


class ExploreProcedure[T_takes: Entry, T_gives: Entry](ABC):
    takes_entrytype: type[T_takes]
    gives_entrytype: type[T_gives]
    relation_type: type[Relation]  # Filter on the produced relations
    entry_filter: Optional[EntryFilter[T_gives]]   # Filter on T_gives
    relation_filter: Optional[RelationFilter]

    # Pre-computed SQL snippets
    _opt_relation_filter_clause: sql.Composable
    _opt_entry_filter_clause: sql.Composable

    n_params: ClassVar[int]

    # SQL
    q: sql.Composed
    cursor: pg.Cursor

    def __init__(
        self,
        relationtype: type[Relation],
        relation_filter: Optional[RelationFilter],
        entry_filter: Optional[EntryFilter],
    ):
        # Init
        self.relation_type = relationtype
        self.relation_filter = relation_filter
        self.entry_filter = entry_filter

        if relation_filter is not None:
            #assert relation_filter.relation_type == relationtype
            pass
        if entry_filter is not None:
            #assert entry_filter.entry_type == self.gives_entrytype
            pass

        # Generate the various clauses ahead of time
        match self.relation_filter:
            # TODO filtering
            #case RelationFilterSource():
            #    # Filter
            #    self._opt_relation_filter_clause = sql.SQL("""
            #        AND (rel_src in {src_set})
            #    """).format(
            #        src_set = list(relation_filter.allowed_sources)
            #    )
            case _:
                # No filter
                self._opt_relation_filter_clause = sql.SQL('')

        # Entry clause
        # TODO change to filter entries on the DB level (hard for symmetric relations?)
        #   for symmetric - BOTH recon ids need to be in the set of allowed id types?
        self._opt_entry_filter_clause = self.make_entry_filter_clause()

        # Generate the query except for the formats
        # Clause is prepared with as much formating as possible now, %(reconids)s param will be used when
        # calling only
        self.q = sql.SQL("""
          SELECT *
          FROM {rel_view}
          WHERE {where_clause} {opt_relation_clause} {opt_entry_clause}
          ;
        """).format(
            rel_view = sql.Identifier(f'{self.relation_type.get_table_name()}_v'),
            where_clause = self.make_where_clause(),
            opt_relation_clause = self._opt_relation_filter_clause,
            opt_entry_clause = self._opt_entry_filter_clause
        )

    @abstractmethod
    def make_where_clause(self) -> sql.SQL:
        """ Returns a parametrised SQL string to be used when calling. """
        pass

    @abstractmethod
    def make_entry_filter_clause(self) -> sql.SQL:
        """ Returns an SQL string filtering correctly by recon_id_1, recon_id_2, or both. """
        pass


class ExploreProcedureT1[T_takes: Entry, T_gives: Entry](ExploreProcedure[T_takes, T_gives]):
    # Explore from T1
    takes_entrytype: type[T_takes]
    gives_entrytype: type[T_gives]
    relation_type: type[Relation[T_takes, T_gives]]
    entry_filter: Optional[EntryFilter[T_gives]]
    relation_filter: Optional[RelationFilter[T_takes, T_gives]]

    n_params: ClassVar[int] = 1

    def __init__(
        self,
        relationtype: type[Relation[T_takes, T_gives]],
        relation_filter: Optional[RelationFilter[T_takes, T_gives]],
        entry_filter: Optional[EntryFilter[T_gives]]
    ):
        assert not relationtype.symmetric
        self.takes_entrytype = relationtype.source_entrytype
        self.produces_entrytype = relationtype.target_entrytype
        self.relation = relationtype

        super().__init__(relationtype, relation_filter, entry_filter)

    def make_where_clause(self) -> sql.SQL:
        return sql.SQL("recon_id_1 = %s")

    def make_entry_filter_clause(self) -> sql.SQL:
        # TODO
        return sql.SQL('')


class ExploreProcedureT2[T_takes: Entry, T_gives: Entry](ExploreProcedure[T_takes, T_gives]):
    # Explore from T2
    takes_entrytype: type[T_takes]
    gives_entrytype: type[T_gives]
    relation_type: type[Relation[T_gives, T_takes]]
    entry_filter: Optional[EntryFilter[T_gives]]
    relation_filter: Optional[RelationFilter[T_gives, T_takes]]

    n_params: ClassVar[int] = 1

    def __init__(
        self,
        relationtype: type[Relation[T_gives, T_takes]],
        relation_filter: Optional[RelationFilter[T_gives, T_takes]],
        entry_filter: Optional[EntryFilter[T_gives]]
    ):
        assert not relationtype.symmetric
        self.takes_entrytype = relationtype.target_entrytype
        self.produces_entrytype = relationtype.source_entrytype
        self.relation = relationtype

        super().__init__(relationtype, relation_filter, entry_filter)

    def make_where_clause(self) -> sql.SQL:
        return sql.SQL("recon_id_2 = %s")

    def make_entry_filter_clause(self) -> sql.SQL:
        # TODO
        return sql.SQL('')


class ExploreProcedureSym[T: Entry](ExploreProcedure[T, T]):
    # Explore from T
    takes_entrytype: type[T]
    gives_entrytype: type[T]
    relation_type: type[Relation[T, T]]
    entry_filter: Optional[EntryFilter[T]]
    relation_filter: Optional[RelationFilter[T, T]]

    n_params: ClassVar[int] = 2

    def __init__(
        self,
        relationtype: type[Relation[T, T]],
        relation_filter: Optional[RelationFilter[T, T]],
        entry_filter: Optional[EntryFilter[T]]
    ):
        assert relationtype.symmetric
        self.takes_entrytype = relationtype.source_entrytype
        self.produces_entrytype = relationtype.target_entrytype
        self.relation = relationtype

        super().__init__(relationtype, relation_filter, entry_filter)

    def make_where_clause(self) -> sql.SQL:
        return sql.SQL("(recon_id_1 = %s) OR (recon_id_2 = %s)")

    def make_entry_filter_clause(self) -> sql.SQL:
        # TODO
        return sql.SQL('')



from typing import Optional

from chemrecon.schema import Entry, FeatureEnum
from chemrecon.schema.db_object import Column, col_recon_id


class MolStructure(Entry):
    """ Abstract representation of a structure in SMILES format.
    """
    # Attributes
    smiles: str  #: The canonical SMILES identifier.
    std_feats: Optional[list[FeatureEnum]]  #: The features w.r.t which this structure is standardized (see TODO).
    molformula: Optional[str]  #: The sum formula.

    # Database
    entrytype_name = 'MolStructure'
    _table_name = 'MolStructure'
    _columns = [
        col_recon_id,
        Column('smiles', str, index_hash = True),
        Column('std_feats', list[FeatureEnum]),
        Column('molformula', str),
    ]
    _index = [1]

    # Visualisation
    _draw_colour = '00FFFF'

    def __init__(
        self,
        smiles: str,
        std_feats: Optional[list[FeatureEnum]] = None,
        molformula: Optional[str] = None,
        recon_id: Optional[int] = None,
    ):
        super().__init__(recon_id)
        self.smiles = smiles
        self.std_feats = std_feats
        self.molformula = molformula

        # Calculate std. features if none
        if self.std_feats is None:
            from chemrecon import mol_from_smiles
            mol = mol_from_smiles(smiles = self.smiles)
            self.std_feats = {feat.feature_enum for feat in mol.features}

        # Calculate sum formula if not given
        if self.molformula is None:
            # TODO
            pass

    def _vis_str(self) -> str:
        return f'{self.smiles}\n({self._feat_str()})'

    def _feat_str(self) -> str:
        if self.std_feats is None:
            return '?'
        s: str = ''
        s += 'F,' if FeatureEnum.F in self.std_feats else '-,'
        s += 'I,' if FeatureEnum.I in self.std_feats else '-,'
        s += 'C,' if FeatureEnum.C in self.std_feats else '-,'
        s += 'T,' if FeatureEnum.T in self.std_feats else '-,'
        s += 'S,' if FeatureEnum.S in self.std_feats else '-,'
        return s



from typing import Optional

from chemrecon.schema.enums import SourceDatabase
from chemrecon.schema.db_object import Relation, col_src, InverseRelation
from chemrecon.schema.entry_types.reaction import Reaction


# Is A / Has Instance
# ----------------------------------------------------------------------------------------------------------------------
class ReactionIsA(Relation[Reaction, Reaction]):
    """ Hierarchical relation which indicates that a reaction is a member of a class of reactions.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Is a'
    _table_name = 'ReactionIsA'
    symmetric = False
    source_entrytype = Reaction
    target_entrytype = Reaction
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'is a'

class ReactionHasInstance(InverseRelation[Reaction, Reaction]):
    """ Hierarchical relation which indicates that a reaction is a member of a class of reactions.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Inverse
    inverse_main_relation = ReactionIsA

    # Database
    entrytype_name = 'Has Instance'
    _table_name = 'ReactionHasInstance'
    symmetric = False
    source_entrytype = Reaction
    target_entrytype = Reaction
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'has instance'

# Set Inverse
ReactionIsA.has_inverse = ReactionHasInstance

# New / Old id
# ----------------------------------------------------------------------------------------------------------------------
class ReactionHasNewID(Relation[Reaction, Reaction]):
    """ Indicates correspondence between identifiers in different database versions. Can be used to resolve deprecated
        identifiers.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Has New ID'
    _table_name = 'ReactionHasNewID'
    symmetric = False
    source_entrytype = Reaction
    target_entrytype = Reaction
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'new id'

class ReactionHasOldID(InverseRelation[Reaction, Reaction]):
    """ Indicates correspondence between identifiers in different database versions. Can be used to resolve deprecated
        identifiers.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Inverse
    inverse_main_relation = ReactionHasNewID

    # Database
    entrytype_name = 'Has Old ID'
    _table_name = 'ReactionHasOldID'
    symmetric = False
    source_entrytype = Reaction
    target_entrytype = Reaction
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            src: SourceDatabase,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'old id'

# Set Inverse
ReactionHasNewID.has_inverse = ReactionHasOldID



""" Implements a heuristic algorithm to match structures of a ChemReaction to the structures attached to a
    Compound.
"""
from __future__ import annotations

from collections import OrderedDict
from enum import Enum
from typing import Hashable

from chemrecon import Direction, ChemReaction, Mol
from chemrecon.chem.create_mol import mol_from_struct_entry
from chemrecon.entrygraph.entrygraph import EntryGraph
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol
from chemrecon.entrygraph.explore import explore
from chemrecon.entrygraph.scoring import Scorer
from chemrecon.schema import (
    Compound, MolStructure,
)
from chemrecon.schema.relation_types_procedural.compound_select_structure_proceduralrelation import \
    CompoundSelectStructure
from chemrecon.utils import hungarian


class FlipState(Enum):
    NORMAL = 0
    FLIPPED = 1


def match_reactants(
        reaction: ChemReaction,
        compound_entries_lhs: dict[str, list[Compound]],
        compound_entries_rhs: dict[str, list[Compound]],
        consider_n_most_confident_structures: int = 1,
        consider_first_entry_primary: bool = True,
        consider_flipped: bool = False,
        search_depth: int = 2
) -> tuple[dict[Mol, str], dict[Mol, str], FlipState]:
    """ ...
        Compound entries are given as a dict (with key = primary_id), and a list of associated entries, possibly only 1.
        If consider_first_entry_primary is set, for each given compound as a list of entries, the first entry in the
        list will have higher weight in the EntryGraph.
        Each compound can be given with multiple entries (for SBML files, for instance).
        If consider_flipped is set, will also consider the case when LHS and RHS are flipped.
        If look_deep is set, will consider structures not directly associated with the entries, instead using a deeper
        Entrygraph search.
        Also returns whether flipped was best
    """

    # First, find structures using EntryGraphs
    compound_structures_lhs: dict[str, OrderedDict[Mol, float]] = OrderedDict()
    compound_structures_rhs: dict[str, OrderedDict[Mol, float]] = OrderedDict()
    for compound_entries, compound_structures in [
        [compound_entries_lhs, compound_structures_lhs],
        [compound_entries_rhs, compound_structures_rhs]
    ]:
        compound_entries: dict[str, list[Compound]]
        compound_structures: dict[str, OrderedDict[Mol, float]]

        for k, entry_list in compound_entries.items():
            try:
                compound_structures[k] = dict()
                for struct_entry, score in search_structures(entry_list).items():
                    try:
                        compound_structures[k][mol_from_struct_entry(struct_entry)] = score
                    except AttributeError:
                        # Could not generate Mol
                        pass
            except ValueError:
                # No valid initial entries
                compound_structures[k] = dict()

    # Perform matching based on the Mols in the ChemReaction
    match_l: dict[Mol, str]
    match_r: dict[Mol, str]
    match_ll, score_ll = match(
        reaction.get_lhs_templates(),
        compound_structures_lhs,
        consider_n_most_confident_structures = consider_n_most_confident_structures
    )
    match_rr, score_rr = match(
        reaction.get_rhs_templates(),
        compound_structures_rhs,
        consider_n_most_confident_structures = consider_n_most_confident_structures
    )

    # Consider the reaction flipped
    if consider_flipped:
        match_lr, score_lr = match(
            reaction.get_lhs_templates(),
            compound_structures_rhs,
            consider_n_most_confident_structures = consider_n_most_confident_structures
        )
        match_rl, score_rl = match(
            reaction.get_rhs_templates(),
            compound_structures_lhs,
            consider_n_most_confident_structures = consider_n_most_confident_structures
        )

        if (score_lr + score_rl) > (score_ll + score_rr):
            print(f'  -> flipped best ({score_lr:.2f} + {score_rl:.2f}) > ({score_ll:.2f} + {score_rr:.2f})')
            flipstate = FlipState.FLIPPED
            match_l = match_lr
            match_r = match_rl
        else:
            print(f'  -> non-flipped best ({score_lr:.2f} + {score_rl:.2f}) < ({score_ll:.2f} + {score_rr:.2f})')
            flipstate = FlipState.NORMAL
            match_l = match_ll
            match_r = match_rr
    else:
        flipstate = FlipState.NORMAL
        match_l = match_ll
        match_r = match_rr

    # Return chosen matching
    return (match_l, match_r, flipstate)


# Matching alg
def match[T: Hashable](
        reaction_mols: list[Mol],
        compound_mols: dict[T, OrderedDict[Mol, float]],
        consider_n_most_confident_structures: int = 1,
) -> tuple[dict[Mol, T], int]:
    """ Given the mols from the reaction and the compound_mols (with confidence), get the best mapping.
        Type parameter T is the key, probalby str (primary_id).
        Returns the matching, as well as a confidence level (0 to 1).
    """
    similarity: dict[tuple[Mol, T], float] = dict()

    # First, compute the pairwise similarity of all compounds
    for mol in reaction_mols:
        for primary_id, mols_ in compound_mols.items():
            # if len(mols_) == 0:
            #    # No structures given
            #    continue

            sims: list[float] = list()
            for mol_, conf in list(mols_.items())[:consider_n_most_confident_structures]:
                # Get distance
                sims.append(mol.get_similarity(mol_) * conf)  # 1 if identical, approaches 0 as difference increases.

            # Calculate the similarity as the maximum of (sim * conf)
            similarity[(mol, primary_id)] = max(sims) if sims else 0

    # Run the Hungarian algorithm to determine the best matching
    matching = hungarian.max_weight_matching(edges = similarity)

    # Compute the confidence/score of the matching as avg. weight of matched edges
    confidence = sum(similarity[mol, primary_id] for mol, primary_id in matching.items()) / len(reaction_mols)
    return matching, confidence


def search_structures(compound_entries: list[Compound]) -> OrderedDict[MolStructure, float]:
    eg = EntryGraph(initial_entries = set(compound_entries))
    explore(eg, structure_protocol, steps = 4)
    return scorer_structure(eg)


# EntryGraph specifications for search
# ----------------------------------------------------------------------------------------------------------------------
structure_protocol = ExplorationProtocol(
    relation_types = {CompoundSelectStructure}
)
scorer_structure = Scorer[MolStructure](
    score_entry_type = MolStructure
)

# TODO score depending on whether primary id is weighted higher



""" TODO re-implement this later, for now, filters should just be functions
"""

from abc import ABC, abstractmethod
from typing import Callable

from chemrecon.schema.db_object import SourceEntry
from chemrecon.core.id_types import IdentifierType
from chemrecon.schema.db_object import Entry, Relation
from chemrecon.schema.enums import SourceDatabase


class EntryFilter[T: Entry](ABC):
    """ Base class for filters. """
    entry_type: type[Entry]
    allowed_id_types: set[IdentifierType]
    has_id_type_field: bool

    @abstractmethod
    def __call__(self, entry: T) -> bool:
        # Base implementation
        return True

class RelationFilter[T1: Entry, T2: Entry](ABC):
    """ Base class for filters. """
    relation_type: type[Relation[T1, T2]]
    allowed_sources: set[SourceDatabase]
    has_src_field: bool

    def __init__(self, relation_type: type[Relation[T1, T2]]):
        self.relation_type = relation_type

    @abstractmethod
    def __call__(self, relation: Relation[T1, T2]) -> bool:
        # Base implementation
        return True

class EntryFilterIdType[T: SourceEntry](EntryFilter[T]):
    """
    """
    allowed_id_types: set[IdentifierType]
    # TODO

    def __init__(self, allowed_id_types: set[IdentifierType]):
        self.allowed_id_types = allowed_id_types
        raise NotImplementedError()

    def __call__(self, entry: T) -> bool:
        raise NotImplementedError()

# TODO similar to EntryFilterIdType, but for source


class EntryFilterProcedure[T: Entry](EntryFilter):
    """ Allows custom filtering on the application level in addition to type filtering.
    """
    procedure: Callable[[T], bool]

    def __init__(self, filter_proc: Callable[[T], bool]):
        self.procedure = filter_proc
        super().__init__()

    def __call__(self, entry: T) -> bool:
        if super().__call__(entry):
            return self.procedure(entry)
        else:
            return False



class RelationFilterProcedure[T1: Entry, T2: Entry](RelationFilter):
    """ Allows custom filtering on the application level in addition to source
    """
    procedure: Callable[[Relation[T1, T2]], bool]

    def __call__(self, relation: Relation[T1, T2]) -> bool:
        if super().__call__(relation):
            return self.procedure(relation)
        else:
            return False

    def __init__(
        self,
        relation_type: type[Relation[T1, T2]],
        filter_proc: Callable[[Relation[T1, T2]], bool]
    ):
        super().__init__(relation_type)
        self.procedure = filter_proc



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, ClassVar, Optional, OrderedDict

from chemrecon.core.id_types import IdentifierType
from chemrecon.schema.enums import Quality, SourceDatabase


class DatabaseObject(ABC):
    """ Most generic database object, covers both entries and relations.
    """
    entrytype_name: ClassVar[str]
    _table_name: ClassVar[str]

    @classmethod
    def get_table_name(cls) -> str:
        return cls._table_name


class Column:
    """ Wrapper for database columns.
    """
    name: str
    col_type: type
    serial: bool  # If numeric, whether serial
    index_hash: bool  # Whether this should be hashed if used in an index.

    def __init__(self, name: str, col_type: type, serial: bool = False, index_hash: bool = False):
        self.name = name
        self.col_type = col_type
        self.serial = serial
        self.index_hash = index_hash

    def __repr__(self):
        return f'col: {self.name}'


# Entry abstract base class
# --------------------------------------------------------------------------------------------------------------
class Entry(DatabaseObject, ABC):
    """ Generic base class for entries.
    """
    # Attributes

    #: Internal identifier in the *ChemRecon* database.
    #: Normally, entries have a nonnegative `recon_id`, unique to the table.
    #: A negative `recon_id` indicates that the object is 'virtual', that is, it was created by a procedural relation,
    #: and does not exist in the database.
    #: A `recon_id` of `None` indicates that the entry is not stored in the database.
    recon_id: Optional[int]

    # Database
    _columns: ClassVar[list[Column]]  # List of columns. The zeroth (recon_id) will be considered the PK.
    _index: ClassVar[list[int]]  # List of the columns on which to create a (possibly hashed) index.

    # For visualization
    _draw_colour: ClassVar[str]

    def get_columns_with_values(self, include_recon_id = True) -> dict[Column, Any]:
        """ Get the columns of this entry with values. """
        return OrderedDict(
            (c, self.__getattribute__(c.name)) for c in self._columns if
            (c is not col_recon_id) or include_recon_id
        )

    def get_index_columns_with_values(self) -> dict[Column, Any]:
        """ Get the index (primary key) columns of this entry with values. """
        return OrderedDict(
            (c, self.__getattribute__(c.name)) for c in self.get_index_columns()
        )

    # Methods for database interaction
    @classmethod
    def get_columns(cls, include_recon_id = True) -> list[Column]:
        if include_recon_id:
            return cls._columns
        else:
            return cls._columns[1:]  # Assuming recon_id is columns[0]

    @classmethod
    def get_index_indices(cls) -> list[int]:
        return cls._index

    @classmethod
    def get_index_columns(cls) -> list[Column]:
        return [cls._columns[i] for i in cls._index]

    # Misc
    def __init__(self, recon_id: Optional[int] = None):
        super().__init__()
        self.recon_id = recon_id

    def __repr__(self):
        attr_cols: str = ', '.join(
            f'{c.name}: {v}'
            for c, v in self.get_columns_with_values(include_recon_id = False).items()
            if (v is not None) and (v != [])
        )
        return f'<{self._table_name} {self.recon_id or '-'}: {attr_cols}>'

    # Comparison and identity
    def __eq__(self, other: Entry):
        if type(self) is not type(other):
            return False
        elif self.recon_id is not None and other.recon_id is not None:
            # Compare recon_id if applicable
            return self.recon_id == other.recon_id
        else:
            # Else, compare by index columns
            return all(
                self.__getattribute__(col.name) == other.__getattribute__(col.name)
                for col in self.get_index_columns()
            )

    def __hash__(self):
        return tuple(self.__getattribute__(col.name) for col in self.get_index_columns()).__hash__()

    def __lt__(self, other: Entry):
        # Compare by reconid, otherwise by index
        if (self.recon_id is not None) and (other.recon_id is not None):
            return self.recon_id.__lt__(other.recon_id)
        else:
            # Compare by index columns
            return tuple(self.get_columns_with_values().values()) < tuple(
                other.get_columns_with_values().values())

    @classmethod
    def get_supertype_of_id_types(cls) -> Optional[type[IdentifierType]]:
        for c in cls.get_columns():
            if c.name == 'id_type':
                if not issubclass(c.col_type, IdentifierType):
                    raise ValueError('Error in identifier type assignment of column.')
                return c.col_type
        return None

    # Visualisation
    @abstractmethod
    def _vis_str(self) -> str:
        """ Get the string of the primary information for visualization.
        """
        pass

    def _vis_attrs(self) -> dict[str, str]:
        """ Define extra attributes used in visualization.
        """
        return {
            'fillcolor': f'"#{self._draw_colour}"'
        }

    # Encoding (for memoization/caching)
    def encode(self):
        return str(self.recon_id).encode()

    def serialise(self, index_only: bool = False) -> dict:
        d = dict()
        if index_only:
            for col, val in self.get_index_columns_with_values().items():
                d[col.name] = serialise_col(col, val)
        else:
            for col, val in self.get_columns_with_values().items():
                d[col.name] = serialise_col(col, val)
        return d


# Entry as a database stand-in
class SourceEntry(Entry, ABC):
    """ An entry which stands for a database entry in one of the source databases.
    """
    source_id: str
    id_type: Enum

    def _vis_str(self) -> str:
        return f'{self.id_type.name}: {self.source_id}'


# Relation
# --------------------------------------------------------------------------------------------------------------
class Relation[T1, T2](DatabaseObject, ABC):
    """ Generic base class for relations between entries.
    """
    # Attributes
    recon_id_1: Optional[int]  #: Recon ID of source
    recon_id_2: Optional[int]  #: Recon ID of target
    entry_1: Optional[T1]  # Entries on either end. Used on prototypes, and can also be fetched via the db.
    entry_2: Optional[T2]

    # Database
    ignore_generation_limit: ClassVar[bool] = False  # If true, will continue exploring after generation limit
    symmetric: ClassVar[bool]  # If symmetric, will always have recon_id_1 <= recon_id_2

    # Set if this is the main table of another inverse relation
    has_inverse: ClassVar[Optional[type[Relation]]]  #: :meta private:

    _attribute_columns: ClassVar[list[Column]]
    _index: ClassVar[list[int]]  # List of col index in attr list on which to create an index.

    source_entrytype: ClassVar[type[Entry]]  #: :meta private:
    target_entrytype: ClassVar[type[Entry]]  #: :meta private:

    # Methods for database interaction
    @classmethod
    def get_attribute_columns(cls) -> list[Column]:
        # Note: does not include source/target columns
        return cls._attribute_columns

    @classmethod
    def get_columns(cls, include_recon_ids: bool = True) -> list[Column]:
        if include_recon_ids:
            return [col_recon_id_1, col_recon_id_2, *cls._attribute_columns]
        else:
            return cls._attribute_columns

    @classmethod
    def get_index_indices(cls) -> list[int]:
        return cls._index

    @classmethod
    def get_index_columns(cls) -> list[Column]:
        return [col_recon_id_1, col_recon_id_2, *[
            cls._attribute_columns[i]
            for i in cls._index
        ]]

    def get_columns_with_values(self, include_recon_ids: bool = True) -> dict[Column, Any]:
        """ Get the columns of this entry with values. """
        return OrderedDict(
            (c, self.__getattribute__(c.name)) for c in self.get_columns(include_recon_ids)
        )

    def get_index_columns_with_values(self) -> dict[Column, Any]:
        """ Get the index (primary key) columns of this entry with values.
            This always includes `recon_id_1` and `recon_id_2`, and may include attribute columns.
        """
        return OrderedDict(
            (c, self.__getattribute__(c.name)) for c in self.get_index_columns()
        )

    @classmethod
    def get_entry_table_names(cls) -> tuple[str, str]:
        return cls.source_entrytype.get_table_name(), cls.target_entrytype.get_table_name()

    # Misc
    def __init__(
            self,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None,
    ):
        super().__init__()
        self.recon_id_1, self.recon_id_2 = recon_id_1, recon_id_2

    def __repr__(self):
        arrow = '<->' if self.symmetric else '->'
        attr_cols: str = ', '.join(
            f'{c.name}: {v}'
            for c, v in self.get_columns_with_values(include_recon_ids = False).items()
        )
        return f'<{self._table_name} {self.recon_id_1} {arrow} {self.recon_id_2}] {attr_cols}>'

    # Visualisation
    def _vis_attrs(self) -> dict[str, str]:
        """ Define extra attributes used in visualization.
        """
        return dict()

    def _vis_str(self) -> str:
        """ Define extra attributes
        """
        return ''

# Inverse relations
class InverseRelation[T1: Entry, T2: Entry](Relation[T1, T2], ABC):
    """ Represents the inverse of another (main) relation.
        The main relation should have the has_inverse tag set.
        Exists in the database only as views of the main relation.
    """
    inverse_main_relation: ClassVar[type[Relation]]  #: :meta private:

    # Check that attribute columns are the same as for the main relation


# Procedural Relations
class ProceduralRelation[T1: Entry, T2: Entry](Relation[T1, T2], ABC):
    """ Procedural relations are not stored in the database but computed at runtime.
        Other than this, they have the same interface as normal relations.
        Some can be computed only "one-way" (e.g. a compound can be standardized, but not un-standardized)
    """

    @classmethod
    @abstractmethod
    def generate(
            cls,
            take_entry: T1,
    ) -> list[tuple[ProceduralRelation[T1, T2], T2]]:
        """ Given a T1, generate relations from that.
        """
        raise NotImplementedError()


class ProceduralGeneratorError(Exception):
    """ Should be thrown by generator methods.
    """
    pass


# Composed Relations
class ComposedRelation[T1, T2, T_intermediate](Relation[T1, T2], ABC):
    # Composed of two other relations
    rel_type_1: ClassVar[type[Relation]]  #: :meta private:
    rel_type_2: ClassVar[type[Relation]]  #: :meta private:
    intermediate_entrytype: ClassVar[type[Entry]]  #: :meta private:

    # Init based on the given relations and intermediate
    @abstractmethod
    def __init__(
            self,
            rel_1: Relation[T1, T_intermediate],
            rel_2: Relation[T_intermediate, T2],
            intermediate: T_intermediate,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None,
    ):
        """ Init for composed relations should be overriden to set attributes of the composed relation based on the
            attributes of relations and entries used to create it.
        """
        super().__init__(recon_id_1, recon_id_2)

    # Filter functions on the composed edges, which determine whether to create the procedural relations
    @classmethod
    def filter_rel_1(cls, r: Relation[T1, T_intermediate]) -> bool:
        return True

    @classmethod
    def filter_intermediate(cls, e: T_intermediate) -> bool:
        return True

    @classmethod
    def filter_rel_2(cls, r: Relation[T_intermediate, T2]) -> bool:
        return True


# Predefined columns
# ----------------------------------------------------------------------------------------------------------------------
col_recon_id = Column('recon_id', int, serial = True)
col_source_id = Column('source_id', str)
col_source_id_hashed = Column('source_id', str, index_hash = True)
col_name = Column('name', str)
col_src = Column('src', SourceDatabase)
col_quality = Column('quality', Quality)
col_properties = Column('properties', list[str])
col_recon_id_2 = Column('recon_id_2', int)
col_recon_id_1 = Column('recon_id_1', int)
col_score = Column('score', float)


# Util
# ----------------------------------------------------------------------------------------------------------------------
def serialise_col(col: Column, value: Any) -> Any:
    if isinstance(value, Enum):
        return value.name
    else:
        return value



from __future__ import annotations

from typing import Callable
import re

# Lookup lists and dictionaries
# ----------------------------------------------------------------------------------------------------------------------
id_types: list[IdentifierType] = list()
id_type_name_lookup: dict[str, IdentifierType] = dict()
identifiers_org_dict: dict[str, IdentifierType] = dict()


# Types of ID types
# ----------------------------------------------------------------------------------------------------------------------
class IdentifierType:
    """
    Represents a general type of identifier with attributes and methods for standardization,
    recognition, and manipulation.

    This class is designed to encapsulate information about a specific type of identifier,
    including its primary name, alternative names, recognizable patterns, and
    standardization logic.

    The class also registers identifier types in global lookup dictionaries
    to facilitate type recognition and access.
    """
    name: str  #: Primary name of this identifier type.
    shortname: str  #: The name used for the type in the database.
    alt_names: set[str]  #: Alternative names to search for.
    prefixes: set[str]
    suffixes: set[str]
    enum_type: 'IdType'  #: The corresponding Enum value, as present in entries
    stdfunc: Callable[[str], str]  #: Function used to standardize identifiers.
    recogniser: re.Pattern | None  #: Pattern used to recognize identifiers of this type.
    id_org_prefix: str  #: Prefix in identifiers.org
    objectname: str  # name of the assigned object

    def __init__(
            self,
            name: str,
            shortname: str,
            alt_names: set[str] = None,
            prefixes: set[str] = None,
            suffixes: set[str] = None,
            stdfunc: Callable[[str], str] = None,
            recogniser: re.Pattern | None = None,
            objectname: str = None
    ):
        self.name = name
        self.shortname = shortname
        self.alt_names = alt_names
        self.prefixes = prefixes if prefixes else {}
        self.suffixes = suffixes if suffixes else {}
        self.stdfunc = stdfunc
        self.recogniser = recogniser
        self.objectname = objectname

        # Make identifiers org link
        # TODO doesn't always work (mnx?)
        self.identifiers_org_prefix = ''
        for prefix in self.prefixes:
            if prefix.startswith('https://identifiers.org/') or prefix.startswith('http://identifiers.org/'):
                identifiers_org_dict[prefix] = self
                self.identifiers_org_prefix = prefix

        # Register in lookup
        id_types.append(self)

        # Register names
        id_type_name_lookup[self.name] = self
        id_type_name_lookup[self.shortname] = self
        for alt_name in self.alt_names:
            id_type_name_lookup[alt_name] = self

    def __repr__(self):
        return self.objectname

    def __str__(self):
        return self.name

    def __hash__(self):
        return self.shortname.__hash__()

    def std_identifier(self, s: str) -> str:
        """ Standardize a given identifier of this type. """
        if self.stdfunc:
            return self.stdfunc(s)
        else:
            return str(s)

    def trim(self, s: str) -> str:
        """ Remove pre- and suffixes of the string, including identifiers.org urls. """
        s = s.strip()
        for prefix in self.prefixes:
            s = s.removeprefix(prefix)
        for suffix in self.suffixes:
            s = s.removesuffix(suffix)
        return s


# Register subtypes
class IdentifierTypeCompound(IdentifierType):
    enum_type: 'IdTypeCompoundEnum'


class IdentifierTypeStructureRepresentation(IdentifierType):
    enum_type: 'IdTypeStructureRepresentationEnum'


class IdentifierTypeReaction(IdentifierType):
    enum_type: 'IdTypeReactionEnum'


class IdentifierTypeEnzyme(IdentifierType):
    enum_type: 'IdTypeEnzymeEnum'


class IdentifierTypeAAM(IdentifierType):
    enum_type: 'IdTypeAAMEnum'


# Standardisation functions
# ----------------------------------------------------------------------------------------------------------------------
# BiGG
_compartment_suffixes = [
    '_e', '_ex', '_c', '_p', '_m', '_x', '_b',
    '_E', '_EX', '_C', '_P', '_M', '_X', '_B'
]


def _std_bigg(s: str) -> str:
    """ Create a universal BIGG id without the compartment suffix"""
    # TODO double underscore is necessary (indicates stereo - "phe__L" is L-Phenylalanine)?
    s_ = s.strip()
    for suffix in _compartment_suffixes:
        s_ = s_.removesuffix(suffix)
    s_ = s_.replace('_DASH', '')  # Fix weird 'DASH' notation, i.e. M_12ppd_DASH_S -> M_12ppd__S
    s_ = s_.replace('__', '_')  # Hacky fix for double underscore notation
    s_ = s_.removeprefix('m_')
    if not s_.startswith('M_'):
        s_ = f'M_{s_}'

    if s_.isupper():
        # Fix upper case
        s_ = f'M_{s_.removeprefix("M_").lower()}'

    return s_


def _std_bigg_r(s: str) -> str:
    s_ = s.strip()
    for suffix in _compartment_suffixes:
        s_ = s_.removesuffix(suffix)
    s_ = s_.replace('_DASH', '').replace('__', '_').removeprefix('r_')
    if not s_.startswith('R_'):
        s_ = f'R_{s_}'
    if s_.isupper():
        # Fix upper case
        s_ = f'R_{s_.removeprefix("M_").lower()}'

    # Replace LPAREN, RPAREN notation in names
    s_ = s_.replace('LPAREN_', '_').replace('RPAREN_', '_')

    return s_


def _std_chebi(chebi: str) -> str:
    """ Chebi IDs should be prefixed with CHEBI: """
    if not chebi:
        raise ValueError(f'Received empty CHEBI')
    chebi_new = chebi.strip()
    if chebi_new.startswith('http://identifiers.org/chebi/CHEBI'):
        return chebi_new.removeprefix('http://identifiers.org/chebi/')
    if chebi_new.startswith('https://identifiers.org/chebi/CHEBI'):
        return chebi_new.removeprefix('https://identifiers.org/chebi/')

    if chebi_new.startswith('CHEBI:'):
        return chebi_new
    else:
        return f'CHEBI:{chebi_new}'


# C_NAME
def _standardise_cname(cname: str) -> str:
    """ Lowercase, etc """
    return cname.lower().strip()


# Compound Identifier Types
# ---------------------------------------------------------------------------------------------------------------------

# Unknown
C_UNKNOWN = IdentifierTypeCompound(
    name = 'Unknown Compound',
    shortname = 'unknown_c',
    alt_names = set(),
    prefixes = set(),
    objectname = 'C_UNKNOWN'
)

# Name
C_NAME = IdentifierTypeCompound(
    name = 'Compound Name',
    shortname = 'cname',
    alt_names = set(),
    prefixes = set(),
    objectname = 'C_NAME',
)

# MetaNetX
C_MNX = IdentifierTypeCompound(
    name = 'MetaNetX',
    shortname = 'mnx',
    alt_names = {'metanetx', 'mnx'},
    prefixes = {
        'http://identifiers.org/metanetx.chemical/',
        'https://identifiers.org/metanetx.chemical/',
    },
    recogniser = re.compile(r'^(MNXM\d+|MNX\d+|BIOMASS|WATER)$'),
    objectname = 'C_MNX',
)

# BiGG
C_BIGG = IdentifierTypeCompound(
    name = 'BiGG',
    shortname = 'bigg',
    alt_names = {'bigg', 'biggM', 'bigg.metabolite'},
    prefixes = {
        'http://bigg.ucsd.edu/models/universal/metabolites/',
        'http://identifiers.org/bigg.metabolite/',
        'https://identifiers.org/bigg.metabolite/',
        'bigg.metabolite:'
    },
    suffixes = {
        '_e', '_ex', '_c', '_p', '_m', '_x', '_b'
    },
    stdfunc = _std_bigg,
    recogniser = re.compile(r'^[a-z_A-Z0-9]+$'),
    objectname = 'C_BIGG',
)  #: asd

# PubChem CID
C_PUBCHEM = IdentifierTypeCompound(
    name = 'PubChem CiD',
    shortname = 'pubchem_cid',
    alt_names = {'pubchem', 'pc_cid', 'cid'},
    prefixes = {
        'http://identifiers.org/pubchem.compound/',
        'https://identifiers.org/pubchem.compound/'
    },
    recogniser = re.compile(r'^\d+$'),
    objectname = 'C_PUBCHEM',
)

# KEGG Compound
C_KEGG = IdentifierTypeCompound(
    name = 'KEGG',
    shortname = 'kegg',
    alt_names = {'kegg', 'keggC', 'kegg.compound', 'KEGG COMPOUND', 'KEGG', 'KEGG COMPOUND accession'},
    prefixes = {
        'http://identifiers.org/kegg.compound/',
        'https://identifiers.org/kegg.compound/'
    },
    recogniser = re.compile(r'^C\d+$'),
    objectname = 'C_KEGG',
)

# ChEBI
C_CHEBI = IdentifierTypeCompound(
    name = 'ChEBI',
    shortname = 'chebi',
    alt_names = {'chebi', 'CHEBI', 'ChEBI'},
    prefixes = {
        'http://identifiers.org/chebi/',
        'https://identifiers.org/chebi/'
    },
    recogniser = re.compile(r'^CHEBI:\d+$'),
    stdfunc = _std_chebi,
    objectname = 'C_CHEBI',
)

# ECMDB
C_ECMDB = IdentifierTypeCompound(
    name = 'ECMDB',
    shortname = 'ecmdb',
    alt_names = {'ecmdb'},
    recogniser = re.compile(r'^ECMDB\d+$'),
    objectname = 'C_ECMDB',
)

# InChI Key
C_INCHIKEY = IdentifierTypeCompound(
    name = 'InChI key',
    shortname = 'inchi_key',
    alt_names = {'inchi_key', 'inchikey'},
    prefixes = {
        'http://identifiers.org/inchikey/',
        'https://identifiers.org/inchikey/'
    },
    recogniser = re.compile(r'^[A-Z]{14}-[A-Z]{10}(-[A-Z])?$'),
    objectname = 'C_INCHIKEY',
)

# SLM
C_SLM = IdentifierTypeCompound(
    name = 'SwissLipids',
    shortname = 'slm',
    alt_names = {'slm', 'SLM'},
    prefixes = {
        'http://identifiers.org/slm/',
        'https://identifiers.org/slm/'
    },
    recogniser = re.compile(r'^SLM:\d+$'),  # TODO check?
    objectname = 'C_SLM',
)

# LipidMaps
C_LIPIDMAPS = IdentifierTypeCompound(
    name = 'LipidMaps',
    shortname = 'lipidmapsm',
    alt_names = {'LipidMapsM', 'lipidmapsM', 'lipidmaps', 'LIPID MAPS'},
    prefixes = {
        'http://identifiers.org/lipidmaps/',
        'https://identifiers.org/lipidmaps/'
    },
    recogniser = re.compile(r'^LM(FA|GL|GP|SP|ST|PR|SL|PK)[0-9]{4}([0-9a-zA-Z]{4,6})?$'),
    objectname = 'C_LIPIDMAPS',
)

# MetaCyc Compound
C_METACYC = IdentifierTypeCompound(
    name = 'MetaCyc Compound',
    shortname = 'metacyc',
    alt_names = {'metacycM', 'metacyc.compound'},
    prefixes = {
        'http://identifiers.org/metacyc.compound/',
        'https://identifiers.org/metacyc.compound/'
    },
    recogniser = re.compile(r'^[A-Za-z0-9+_.%-:]+$'),
    objectname = 'C_METACYC',
)

# EnviPath (Can be both compound and reaction?)
C_ENVIPATH = IdentifierTypeCompound(
    name = 'enviPath',
    shortname = 'envipath',
    alt_names = {'envipathM'},
    prefixes = {
        'http://identifiers.org/envipath/',
        'https://identifiers.org/envipath/'
    },
    recogniser = re.compile(
        r'^[\w^_]{8}-[\w^_]{4}-[\w^_]{4}-[\w^_]{4}-[\w^_]{12}/[\w-]+/[\w^_]{8}-'
        r'[\w^_]{4}-[\w^_]{4}-[\w^_]{4}-[\w^_]{12}$'
    ),
    objectname = 'C_ENVIPATH',
)

# Seed Compounds
C_SEED = IdentifierTypeCompound(
    name = 'Seed Compound',
    shortname = 'seed',
    alt_names = {'seedM', 'seed.compound'},
    prefixes = {
        'http://identifiers.org/seed.compound/',
        'https://identifiers.org/seed.compound/'
    },
    recogniser = re.compile(r'^cpd\d+$'),
    objectname = 'C_SEED',
)

# Sabio-RK compound
C_SABIORK = IdentifierTypeCompound(
    name = 'Sabio-RK Compound',
    shortname = 'sabiork',
    alt_names = {'sabiork', 'sabiorkM', 'sabiork.compound'},
    prefixes = {
        'http://identifiers.org/sabiork.compound/',
        'https://identifiers.org/sabiork.compound/'
    },
    recogniser = re.compile(r'^\d+$'),
    objectname = 'C_SABIORK',
)

# HMDB
C_HMDB = IdentifierTypeCompound(
    name = 'HMDB',
    shortname = 'hmdb',
    alt_names = {'hmdb', 'HMDB'},
    prefixes = {
        'http://identifiers.org/hmdb/',
        'https://identifiers.org/hmdb/'
    },
    objectname = 'C_HMDB',
)

# Reactome
C_REACTOME = IdentifierTypeCompound(
    name = 'Reactome',
    shortname = 'reactome',
    alt_names = {'reactomeM', 'reactome.compound'},
    prefixes = {
        'http://identifiers.org/reactome/',
        'https://identifiers.org/reactome/'
    },
    recogniser = re.compile(r'(^R-[A-Z]{3}-\d+(-\d+)?(\.\d+)?$)|(^REACT_\d+(\.\d+)?$)'),
    objectname = 'C_REACTOME',
)

# PDBe
C_PDBE = IdentifierTypeCompound(
    name = 'PBDE Compound',
    shortname = 'pdbe',
    alt_names = {'PDBeChem'},
    objectname = 'C_PDBE',
)

C_BIOCYC = IdentifierTypeCompound(
    name = 'BioCyc Compound',
    shortname = 'biocyc',
    alt_names = {'biocyc'},
    prefixes = {
        'https://identifiers.org/biocyc/',
        'http://identifiers.org/biocyc/'
    },
    recogniser = re.compile(r'^[A-Z-0-9]+(:)?[A-Za-z0-9+_.%-:]+$'),
    objectname = 'C_BIOCYC',
)

# MetaMDB
C_METAMDB = IdentifierTypeCompound(
    name = 'MetaMDB Compound',
    shortname = 'metamdb_c',
    alt_names = {'metamdb_c'},
    objectname = 'C_METAMDB',
)

C_BRENDA = IdentifierTypeCompound(
    name = 'Brenda Compound',
    shortname = 'brenda_c',
    alt_names = {'brenda_c'},
    objectname = 'C_BRENDA',
)

# TODO ChemSpider

# Structure representation Identifier Types
# ---------------------------------------------------------------------------------------------------------------------

S_UNKNOWN = IdentifierTypeStructureRepresentation(
    name = 'Unknown structure',
    shortname = 'unknown_s',
    alt_names = set(),
    prefixes = set(),
    objectname = 'S_UNKNOWN',
)

# S_SMILES
S_SMILES = IdentifierTypeStructureRepresentation(
    name = 'S_SMILES',
    shortname = 'smiles',
    alt_names = {'smiles'},
    objectname = 'S_SMILES',
)

# InChI
S_INCHI = IdentifierTypeStructureRepresentation(
    name = 'InChI',
    shortname = 'inchi',
    alt_names = {'inchi'},
    objectname = 'S_INCHI',
)

# MolFile
S_MOLFILE = IdentifierTypeStructureRepresentation(
    name = 'Molfile',
    shortname = 'molfile',
    alt_names = {'molfile'},
    objectname = 'S_MOLFILE',
)

# S_GML
S_GML = IdentifierTypeStructureRepresentation(
    name = 'GML',
    shortname = 'gml',
    alt_names = {'gml'},
    objectname = 'S_GML',
)

# Reaction Identifier Types
# ---------------------------------------------------------------------------------------------------------------------
R_UNKNOWN = IdentifierTypeReaction(
    name = 'Unknown Reaction',
    shortname = 'unknown_r',
    alt_names = set(),
    prefixes = set(),
    objectname = 'R_UNKNOWN',
)

R_NAME = IdentifierTypeReaction(
    name = 'Reaction Name',
    shortname = 'rname',
    alt_names = {'rname'},
    objectname = 'R_NAME',
)

R_MNX = IdentifierTypeReaction(
    name = 'MetaNetX Reaction',
    shortname = 'mnx_r',
    alt_names = {'mnxr'},
    prefixes = {
        'https://identifiers.org/metanetx.reaction/',
        'http://identifiers.org/metanetx.reaction/'
    },
    recogniser = re.compile(r'^(MNXR\d+|EMPTY)$'),
    objectname = 'R_MNX',
)

R_METACYC = IdentifierTypeReaction(
    name = 'MetaCyc Reaction',
    shortname = 'metacyc_r',
    alt_names = {'metacycr', 'metacycR', 'metacyc.reaction'},
    prefixes = {
        'https://identifiers.org/metacyc.reaction/',
        'http://identifiers.org/metacyc.reaction/'
    },
    recogniser = re.compile(r'^[A-Za-z0-9+_.%-:]+$'),
    objectname = 'R_METACYC',

)

R_BIGG = IdentifierTypeReaction(
    name = 'BiGG Reaction',
    shortname = 'bigg_r',
    alt_names = {'biggr', 'bigg_r', 'biggR', 'bigg.reaction'},
    prefixes = {
        'https://identifiers.org/bigg.reaction/',
        'http://identifiers.org/bigg.reaction'
    },
    recogniser = re.compile(r'^[a-z_A-Z0-9]+$'),
    stdfunc = _std_bigg_r,
    objectname = 'R_BIGG',
)

R_SEED = IdentifierTypeReaction(
    name = 'SEED Reaction',
    shortname = 'seed_r',
    alt_names = {'seed_r', 'seedR', 'seed.reaction'},
    recogniser = re.compile(r'^rxn\d+$'),
    objectname = 'R_SEED',
)

R_KEGG = IdentifierTypeReaction(
    name = 'KEGG Reaction',
    shortname = 'kegg_r',
    alt_names = {'kegg_r', 'keggR', 'kegg.reaction'},
    prefixes = {
        'https://identifiers.org/seed.reaction/',
        'http://identifiers.org/seed.reaction'
    },
    recogniser = re.compile(r'^R\d+$'),
    objectname = 'R_KEGG',
)

R_RHEA = IdentifierTypeReaction(
    name = 'RHEA Reaction',
    shortname = 'rhea_r',
    alt_names = {'rheaR', 'rhea'},
    prefixes = {
        'https://identifiers.org/rhea/',
        'http://identifiers.org/rhea/'
    },
    recogniser = re.compile(r'^\d{5}$'),
    objectname = 'R_RHEA',
)

R_SABIORK = IdentifierTypeReaction(
    name = 'Sabio RK Reaction',
    shortname = 'sabiork_r',
    alt_names = {'sabiorkR', 'sabiork.reaction'},
    prefixes = {
        'https://identifiers.org/sabiork.reaction/',
        'http://identifiers.org/sabiork.reaction'
    },
    recogniser = re.compile(r'^\d+$'),
    objectname = 'R_SABIORK',
)

R_METAMDB = IdentifierTypeReaction(
    name = 'MetaMDB reaction',
    shortname = 'metamdb_r',
    alt_names = {'metamdb_r'},
    objectname = 'R_METAMDB',
)

R_MCSA = IdentifierTypeReaction(
    name = 'MCSA reaction',
    shortname = 'mcsa_r',
    alt_names = {'mcsa_r'},
    objectname = 'R_MCSA',
)

R_BRENDA = IdentifierTypeReaction(
    name = 'Brenda Reaction',
    shortname = 'brenda_r',
    alt_names = {'brenda_r'},
    objectname = 'R_BRENDA',
)

# Enzyme Identifier Types
# --------------------------------------------------------------------------------------------------------------
E_UNKNOWN = IdentifierTypeEnzyme(
    name = 'Unknown enzyme',
    shortname = 'unknown_e',
    alt_names = set(),
    prefixes = set(),
    objectname = 'E_UNKNOWN',
)

E_NAME = IdentifierTypeEnzyme(
    name = 'Enzyme Name',
    shortname = 'ename',
    alt_names = {'ename'},
    objectname = 'E_NAME',
)

E_EC = IdentifierTypeEnzyme(
    name = 'EC',
    shortname = 'ec',
    alt_names = {'ec'},
    prefixes = {
        'http://identifiers.org/ec-code/',
        'https://identifiers.org/ec-code/'
    },
    recogniser = re.compile(r'^\d+\.-\.-\.-|\d+\.\d+\.-\.-|\d+\.\d+\.\d+\.-|\d+\.\d+\.\d+\.(n)?\d+$'),
    objectname = 'E_EC',
)

# Atom-to-Atom Map Identifier Types
# --------------------------------------------------------------------------------------------------------------
A_UNKNOWN = IdentifierTypeAAM(
    name = 'Unknown AAM',
    shortname = 'unknown_a',
    alt_names = set(),
    prefixes = set(),
    objectname = 'A_UNKNOWN',
)

A_REACTIONSMILES = IdentifierTypeAAM(
    name = 'Reaction Smiles',
    shortname = 'rsmiles',
    alt_names = {'rsmiles'},
    objectname = 'A_REACTIONSMILES',
)

A_RXN = IdentifierTypeAAM(
    name = 'RXN',
    shortname = 'rxn',
    alt_names = {'rxn'},
    objectname = 'A_RXN',
)

A_GML_RULE = IdentifierTypeAAM(
    name = 'S_GML rule',
    shortname = 'gml_rule',
    alt_names = {'gml_rule'},
    objectname = 'A_GML_RULE',
)

# Lists of ID types
# --------------------------------------------------------------------------------------------------------------
id_types_compound: list[IdentifierTypeCompound] = [
    C_MNX, C_BIGG, C_PUBCHEM, C_KEGG, C_CHEBI, C_ECMDB, C_INCHIKEY, C_SLM, C_LIPIDMAPS, C_METACYC, C_ENVIPATH,
    C_SEED, C_SABIORK, C_HMDB, C_REACTOME, C_BIOCYC, C_METAMDB,
]
id_types_structure_representation: list[IdentifierTypeStructureRepresentation] = [
    S_SMILES, S_INCHI, S_MOLFILE, S_GML,
]
id_types_reaction: list[IdentifierTypeReaction] = [
    R_BIGG, R_MNX, R_METACYC, R_SEED, R_KEGG, R_RHEA, R_SABIORK, R_METAMDB, R_MCSA,
]
id_types_enzyme: list[IdentifierTypeEnzyme] = [
    E_EC
]
id_types_aam: list[IdentifierTypeAAM] = [
    A_UNKNOWN, A_REACTIONSMILES, A_RXN, A_GML_RULE
]



""" Protocols define how to explore an entrygraph.
"""

from __future__ import annotations

from typing import Optional, Callable

from chemrecon.entrygraph.explore_procedure import (
    ExploreProcedure, ExploreProcedureT1, ExploreProcedureT2,
    ExploreProcedureSym,
)
from chemrecon.entrygraph.filter import EntryFilter, RelationFilter, EntryFilterProcedure, RelationFilterProcedure
from chemrecon.schema import Entry, ProceduralRelation, Relation, ComposedRelation
from chemrecon.schema.direction import Direction


# Protocol specification
class ExplorationProtocol():
    """
    Represents a protocol for exploring relationships and entries within a graph structure.

    This class is designed to define and manage the exploration of entries and relationships in a graph-like
    data model. Users can specify sets of entry and relationship types, as well as define filters and
    procedures for exploration.
    """
    relation_types: set[tuple[type[Relation], Direction]]  #: List of relation types to traverse.
    entry_types: set[type[Entry]]  #: List of entry types involved in the specified relations
    relation_types_terminal: set[tuple[type[Relation], Direction]]  # Will not be explored, only added.
    entry_filters: dict[type[Entry], EntryFilter]  #: Optional filters for each type of entry.
    relation_filters: dict[type[Relation], RelationFilter]  #: Ditto, for relations.

    # Explore procedures
    explore_procedures: list[ExploreProcedure]
    post_explore_procedures: list[ExploreProcedure]
    transitive_subprocedures: dict[type[Relation], tuple[ExploreProcedure, ExploreProcedure]]

    def __init__(
            self,
            relation_types: set[type[Relation] | tuple[type[Relation], Direction]],
            relation_types_terminal: Optional[set[type[Relation]]] = None,
            entry_filters: Optional[dict[type[Entry], Callable[[Entry], bool]]] = None,
            relation_filters: Optional[dict[type[Relation], Callable[[Relation], bool]]] = None,
    ):
        """ Specify an exploration protocol.

        :param relation_types: A set of relation typess to explore. Optionally, (Relation, Direction) tuples can be
                passed to specify in which direction each relation can be traversed. By default, only the forwards
                direction is traversed. For symmetric relations, only the Direction.SYMMETRIC value is allowed.
        :type relation_types: set[type[Relation] | tuple[type[Relation], Direction]],
        :param relation_types_terminal: A set of relation types which are not used for expanding the graph, but are
                added if both endpoints were already found. Here, directionality is not specified.
        :type relation_types_terminal: Optional[set[type[Relation] | tuple[type[Relation], Direction]]]
        :param entry_filters: An optional dictionary of filters for each entry type. Filters should  be a function which
                accepts an entry, and returns False if the entry should not be included by the protocol.
        :type entry_filters: Optional[dict[type[Entry], Callable[[Entry], bool]]]
        :param relation_filters: Ditto, for relations.
        :type relation_filters: Optional[dict[type[Relation], Callable[[Relation], bool]]]
        """
        # Process relation types (assign forward/symmetric direction by default)
        self.relation_types = set()
        for item in relation_types:
            match item:
                case (reltype, direction):
                    # Given with direction
                    self.relation_types.add((reltype, direction))
                case reltype:
                    # Set forwards/symmetric direction by default
                    if reltype.symmetric:
                        self.relation_types.add((reltype, Direction.SYMMETRIC))
                    else:
                        self.relation_types.add((reltype, Direction.FORWARDS))

        self.relation_types_terminal = set()
        if relation_types_terminal is not None:
            for reltype in relation_types_terminal:
                # Set forwards/symmetric direction by default
                if reltype.symmetric:
                    self.relation_types_terminal.add((reltype, Direction.SYMMETRIC))
                else:
                    self.relation_types_terminal.add((reltype, Direction.BOTH))

        # Process filters
        self.entry_filters = dict()
        self.relation_filters = dict()
        if entry_filters is not None:
            for e_type, e_filter in entry_filters.items():
                self.entry_filters[e_type] = EntryFilterProcedure(
                    filter_proc = e_filter
                )
        if relation_filters is not None:
            for r_type, r_filter in relation_filters.items():
                self.relation_filters[r_type] = RelationFilterProcedure(
                    relation_type = r_type,
                    filter_proc = r_filter
                )

        self.entry_filters = (entry_filters) if entry_filters else dict()
        self.relation_filters: dict[type[Relation], Callable[[Relation], bool]] \
            = relation_filters if relation_filters else dict()

        # Set entry types based on the relations
        self.entry_types = set()
        for r, direction in self.relation_types:
            self.entry_types.add(r.source_entrytype)
            self.entry_types.add(r.target_entrytype)

        # List of finished ExplorationProcedure objects to call when exploring
        self.explore_procedures: list[ExploreProcedure] = list()
        self.post_explore_procedures: list[ExploreProcedure] = list()
        self.transitive_subprocedures: dict[type[Relation], tuple[ExploreProcedure, ExploreProcedure]] = dict()

        # Sanity checks of the given input
        for rtype, direction in self.relation_types:
            if rtype.source_entrytype not in self.entry_types:
                raise ValueError(f'Relation {rtype} refers to entry types not allowed in the graph.')
            if rtype.target_entrytype not in self.entry_types:
                raise ValueError(f'Relation {rtype} refers to entry types not allowed in the graph.')

        if len(self.relation_types_terminal.intersection(self.relation_types)) > 0:
            raise ValueError('Terminal relation types must not intersect relation types.')

        # Add 'None' for filters when not specified
        for entrytype in (
                self.entry_types.difference(self.entry_filters.keys())
        ):
            self.entry_filters[entrytype] = None
        for reltype in ({rel for rel, _ in self.relation_types}
                .union({rel for rel, _ in self.relation_types_terminal})
                .difference(self.relation_filters.keys())
        ):
            self.relation_filters[reltype] = None

        # Establish a series of exploration procedures
        # TODO accept no arguments, and assume (forwards) or (symmetric)

        for given_rels, proc_list in [
            [self.relation_types, self.explore_procedures],
            [self.relation_types_terminal, self.post_explore_procedures]
        ]:
            for rel_type, direction in given_rels:
                rel_type: type[Relation]
                if issubclass(rel_type, ProceduralRelation):
                    if direction not in {Direction.FORWARDS, Direction.SYMMETRIC}:
                        raise ValueError('Procedural relations can only be explored forwards.')

                match direction:
                    case Direction.FORWARDS:
                        # Forwards, explore from T1
                        assert not rel_type.symmetric
                        proc_list.append(ExploreProcedureT1(
                            relationtype = rel_type,
                            relation_filter = self.relation_filters[rel_type],
                            entry_filter = self.entry_filters[rel_type.target_entrytype]
                        ))
                    case Direction.BACKWARDS:
                        # Backwards, explore from T2
                        assert not rel_type.symmetric
                        proc_list.append(ExploreProcedureT2(
                            relationtype = rel_type,
                            relation_filter = self.relation_filters[rel_type],
                            entry_filter = self.entry_filters[rel_type.source_entrytype]
                        ))
                    case Direction.BOTH:
                        # Both
                        assert not rel_type.symmetric
                        proc_list.append(ExploreProcedureT1(
                            relationtype = rel_type,
                            relation_filter = self.relation_filters[rel_type],
                            entry_filter = self.entry_filters[rel_type.target_entrytype]
                        ))
                        proc_list.append(ExploreProcedureT2(
                            relationtype = rel_type,
                            relation_filter = self.relation_filters[rel_type],
                            entry_filter = self.entry_filters[rel_type.source_entrytype]
                        ))
                    case Direction.SYMMETRIC:
                        # Symmetric
                        assert rel_type.symmetric
                        proc_list.append(ExploreProcedureSym(
                            relationtype = rel_type,
                            relation_filter = self.relation_filters[rel_type],
                            entry_filter = self.entry_filters[rel_type.target_entrytype]
                        ))

                # If transitive, add sub procedures
                if issubclass(rel_type, ComposedRelation):
                    self.transitive_subprocedures[rel_type] = (
                        ExploreProcedureT1(
                            relationtype = rel_type.rel_type_1,
                            relation_filter = None, entry_filter = None
                        ),
                        ExploreProcedureT1(
                            relationtype = rel_type.rel_type_2,
                            relation_filter = None, entry_filter = None
                        )
                    )

        # done
        pass



""" Defines the ontology entrytypes
"""

from __future__ import annotations

from typing import Optional


# Define the base class and possible properties of the ontology relations
# ----------------------------------------------------------------------------------------------------------------------
class OntologyType:
    name:           str
    description:    str
    synonyms:       dict[str, str]      # For each source, what this relationship is called

    # Applies to compounds / reactions
    for_comp:       bool
    for_reac:       bool
    for_enzy:       bool

    # Relational properties
    symmetric:      bool
    transitive:     bool
    inverse:        Optional[OntologyType]
        # If there is a corresponding inverse relation. Not applicable if symmetric
    is_primary:     bool  # If relation has an inverse, is_primary is true if this relation is shown
    # If no primaries, then both are shown with a dash /
    # Drawing
    draw:           str                 # 'full' or 'arrow'

    enum_type: 'Ontology'

    def __init__(
        self,
        name: str,
        description: str,
        for_comp: bool,
        for_reac: bool,
        for_enzy: bool,
        symmetric: bool,
        transitive: bool = False,
        inverse: OntologyType = None,
        is_primary: bool = False,
        synonyms: dict[str, str] = None,
        draw = 'full'
    ):
        self.name = name
        self.description = description
        self.for_comp = for_comp
        self.for_reac = for_reac
        self.for_enzy = for_enzy
        self.symmetric = symmetric
        self.transitive = transitive
        self.inverse = inverse
        self.is_primary = is_primary
        self.synonyms = synonyms if synonyms else {}
        self.draw = draw



# Define the ontologies
# ----------------------------------------------------------------------------------------------------------------------

IS_A = OntologyType(
    name = 'Is a',
    description = 'Defines a subclass relation.',
    for_comp = True,
    for_reac = True,
    for_enzy = True,
    symmetric = False,
    synonyms = {
        'chebi': 'is_a'
    }
)

HAS_INSTANCE = OntologyType(
    name = 'Has instance',
    description = 'Reverse of is_a relation',
    for_comp = True,
    for_reac = True,
    for_enzy = True,
    symmetric = False,
)
IS_A.inverse = HAS_INSTANCE
HAS_INSTANCE.inverse = IS_A
IS_A.is_primary = True

HAS_PART = OntologyType(
    name = 'Has part',
    description = 'TODO',
    for_comp = True,
    for_reac = False,
    for_enzy = False,
    symmetric = False,
    synonyms = {
        'chebi': 'has_part'
    }
)

TAUTOMER = OntologyType(
    name = 'Tautomer',
    description = 'TODO',
    for_comp = True,
    for_reac = False,
    for_enzy = False,
    symmetric = True,
    transitive = True,
    synonyms = {
        'chebi': 'is_tautomer_of'
    }
)

CONJUCATE_ACID = OntologyType(
    name = 'Conjugate acid',
    description = 'TODO',
    for_comp = True,
    for_reac = False,
    for_enzy = False,
    symmetric = False,
    transitive = False,
    synonyms = {
        'chebi': 'is_conjugate_acid_of'
    }
)

CONJUGATE_BASE = OntologyType(
    name = 'Conjugate base',
    description = 'TODO',
    for_comp = True,
    for_reac = False,
    for_enzy = False,
    symmetric = False,
    transitive = False,
    synonyms = {
        'chebi': 'is_conjugate_base_of'
    }
)
CONJUCATE_ACID.inverse = CONJUGATE_BASE
CONJUGATE_BASE.inverse = CONJUCATE_ACID

STEREOISOMER = OntologyType(
    name = 'Stereoisomer',
    description = 'TODO',
    for_comp = True,
    for_reac = False,
    for_enzy = False,
    symmetric = True,
    transitive = True,
    synonyms = {}
)

ENANTIOMER = OntologyType(
    name = 'Enantiomer',
    description = 'TODO',
    for_comp = True,
    for_reac = False,
    for_enzy = False,
    symmetric = True,
    transitive = True,
    synonyms = {
        'chebi': 'is_enantiomer_of'
    }
)

HAS_NEW_ID = OntologyType(
    name = 'New ID',
    description = 'Is deprecated ID of',
    for_comp = True,
    for_reac = True,
    for_enzy = True,
    symmetric = False,
    transitive = True,
    synonyms = {},
    draw = 'arrow'
)

HAS_OLD_ID = OntologyType(
    name = 'Old ID',
    description = 'Has deprecated ID',
    for_comp = True,
    for_reac = True,
    for_enzy = True,
    symmetric = False,
    transitive = True,
    synonyms = {},
    draw = 'arrow'
)
HAS_NEW_ID.inverse = HAS_OLD_ID
HAS_OLD_ID.inverse = HAS_NEW_ID
HAS_NEW_ID.is_primary = True


# Make list of all ontology relations
# ----------------------------------------------------------------------------------------------------------------------
ontology_relations: list[OntologyType] = [
    IS_A, HAS_INSTANCE, HAS_PART, TAUTOMER, CONJUCATE_ACID, CONJUGATE_BASE, ENANTIOMER, HAS_NEW_ID, HAS_OLD_ID
]

ont_str_dict: dict[str, OntologyType] = dict()
for rel in ontology_relations:
    ont_str_dict[rel.name] = rel
    for syn in rel.synonyms.values():
        ont_str_dict[syn] = rel

def get_relation_type_from_str(s: str) -> OntologyType | None:
    try:
        return ont_str_dict[s]
    except KeyError:
        return None



""" Contains functions to find relations (including procedural) related to an entry.
"""
from typing import DefaultDict

import chemrecon.schema

from chemrecon import Entry, Relation, ProceduralRelation, ComposedRelation
from chemrecon.schema.direction import Direction
from chemrecon.entrygraph.entrygraph import EntryGraph, Edge
from chemrecon.entrygraph.explore import explore
from chemrecon.schema.procedural_relation_entrygraph import ProceduralRelationEG
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol

import chemrecon.connection as globals

# Relation lists for each entry type
# ----------------------------------------------------------------------------------------------------------------------
_entrytype_relation_types: dict[
    type[Entry],
    list[tuple[type[Relation], Direction, ExplorationProtocol]]
] = DefaultDict(list)

for reltype in chemrecon.schema.relationtypes:
    if reltype.symmetric:
        # Symmetric
        protocol_spec = ExplorationProtocol(
            relation_types = {(reltype, Direction.SYMMETRIC)}
        )
        _entrytype_relation_types[reltype.source_entrytype].append(
            (reltype, Direction.SYMMETRIC, protocol_spec)
        )
    else:
        # Not symmetric
        protocol_spec_forward = ExplorationProtocol(
            relation_types = {(reltype, Direction.FORWARDS)},
        )
        _entrytype_relation_types[reltype.source_entrytype].append(
            (reltype, Direction.FORWARDS, protocol_spec_forward)
        )
        if not issubclass(reltype, ProceduralRelation):
            # Add backwards explorer only if not procedural
            protocol_spec_backward = ExplorationProtocol(
                relation_types = {(reltype, Direction.BACKWARDS)},
            )
            _entrytype_relation_types[reltype.target_entrytype].append(
                (reltype, Direction.BACKWARDS, protocol_spec_backward)
            )

# Getter functions
# ----------------------------------------------------------------------------------------------------------------------
def get_relations_from_entry[R: Relation](
        entry: Entry,
        relation_type: type[R],
        get_backward_relations: bool = True
) -> list[tuple[R, Entry]]:
    """ Get relations of a given type with the specified entry as the source.
    """
    # TODO support for get_backward_relations !
    # TODO entrygraph-based method will save procedural relations to the DB!
    # TODO better generics for this!

    if entry.recon_id is None:
        # Get recon_id first
        index_result = globals.handler.get_entry_by_index(entry)
        if not index_result:
            raise ValueError(f'Given entry: {entry} not found in database.')

    res = globals.handler.get_relations_with_entries_by_recon_ids(
        entry_type = type(entry),
        recon_ids = [entry.recon_id],
        relation_type = relation_type
    )
    out_list: list[tuple[Relation, Entry]] = list()
    for target, relations in res[0].items():
        for rel in relations:
            out_list.append((rel, target))
    return out_list

def get_all_relations(
        entry: Entry,
        get_backward_relations: bool = True,
        get_transitive_relations: bool = False,
        get_procedural_relations: bool = False,
        get_procedural_eg_relations: bool = False
) -> list[tuple[Relation, Entry]]:
    """ Get all relations of a given entry.
        Note: Procedural relations cannot be explored backwards.
    """
    # TODO change to entrygraph-based method to save procedural relations to the DB!
    # TODO better generics for this!

    if entry.recon_id is None:
        # Get recon_id first
        index_result = globals.handler.get_entry_by_index(entry)
        if not index_result:
            raise ValueError(f'Given entry: {entry} not found in database.')

    output_list: list[tuple[Relation, Entry]]

    # Create EG spec
    entrytypes: set[type[Entry]] = set()
    relation_types: set[tuple[type[Relation], Direction]] = set()
    for r_type, direction, _ in _entrytype_relation_types[type(entry)]:

        # Skip if not in defined relation types
        if issubclass(r_type, ComposedRelation) and not get_transitive_relations:
            continue
        if issubclass(r_type, ProceduralRelation):
            if not get_procedural_eg_relations: continue
            if issubclass(r_type, ProceduralRelationEG):
                if not get_procedural_eg_relations:
                    continue
        if (not get_backward_relations) and (type(entry) is not r_type.source_entrytype):
            continue

        # Add to specification
        entrytypes.add(r_type.source_entrytype)
        entrytypes.add(r_type.target_entrytype)

        if r_type.symmetric:
            relation_types.add((r_type, Direction.SYMMETRIC))
        else:
            if r_type.source_entrytype == type(entry):
                relation_types.add((r_type, Direction.FORWARDS))
            if get_backward_relations and r_type.target_entrytype == type(entry):
                relation_types.add((r_type, Direction.BACKWARDS))

    protocol_all = ExplorationProtocol(
        relation_types = relation_types,
    )

    # Create an EntryGraph and explore to depth 1
    eg = EntryGraph(
        initial_entries = {entry}
    )
    explore(eg, protocol = protocol_all, steps = 1)

    # Get out-edges of a node (vertex with index 0 is the starting node)
    res = eg.get_out_edges_of_vertex(0)  # list of (Edge, Vertex)
    return [
        (edge.relation, vertex.entry)
        for edge, vertex in res if isinstance(edge, Edge)  # Filter out artificial edges
    ]



""" This script loads and exports connection params.
"""
import os
from typing import Optional


class Params:
    """
    Represents database connection parameters and provides a method to generate a formatted
    connection string compatible with PostgreSQL.
    """
    connection_title: str  # Name of the connection file
    db_name: str
    db_host: str
    db_port: str
    username: str
    password: str

    def __init__(
        self,
        connection_title: str,
        db_name: str,
        db_host: str,
        db_port: str,
        username: str,
        password: str
    ):
        self.connection_title = connection_title
        self.db_name = db_name
        self.db_host = db_host
        self.db_port = db_port
        self.username = username
        self.password = password

    def connection_string(self) -> str:
        """ Returns a connection string compatible with PostgreSQL."""
        return f'postgres://{self.username}:{self.password}@{self.db_host}:{self.db_port}/{self.db_name}'


# Load params
parameter_sets: list[Params] = list()

script_path = os.path.dirname(__file__)
rel_path = 'connection_params/'
params_dir = os.path.join(script_path, rel_path)
for param_file in os.listdir(params_dir):
    try:
        with open(os.path.join(params_dir, param_file)) as f:
            flines = f.readlines()
            param = Params(
                param_file,
                db_name = flines[0].strip(),
                db_host = flines[3].strip(),
                db_port = flines[4].strip(),
                username = flines[1].strip(),
                password = flines[2].strip()
            )
            if len(flines) > 5:
                raise ValueError('Malformed database connection parameters file.')

            parameter_sets.append(param)
    except (KeyError, IndexError):
        print('Invalid connection parameters file.')
        pass

# Set params to find easily

# Test
local_docker_init: Optional[Params] = None
local_docker_dev: Optional[Params] = None
local_docker_pub: Optional[Params] = None

# Production
chemrecon_dev: Optional[Params] = None
chemrecon_pub: Optional[Params] = None

for p in parameter_sets:
    match p.connection_title:
        case 'local_docker_init.dbinfo':
            local_docker_init = p
        case 'local_docker_dev.dbinfo':
            local_docker_dev = p
        case 'local_docker_pub.dbinfo':
            local_docker_pub = p
        case 'chemrecon_dev.dbinfo':
            chemrecon_dev = p
        case 'chemrecon_pub.dbinfo':
            chemrecon_pub = p



from typing import Optional

from chemrecon.schema.db_object import Relation, SourceDatabase, InverseRelation, col_src
from chemrecon.schema.entry_types.enzyme import Enzyme
from chemrecon.schema.entry_types.reaction import Reaction


class ReactionHasEnzyme(Relation[Reaction, Enzyme]):
    """ Relates a reaction entry to the given enzyme which catalyses the reaction.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Database
    entrytype_name = 'Reaction has Enzyme'
    _table_name = 'ReactionHasEnzyme'
    symmetric = False
    source_entrytype = Reaction
    target_entrytype = Enzyme
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
        self,
        src: SourceDatabase = SourceDatabase.unknown,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'catalysed by'

class EnzymeCatalyzesReaction(InverseRelation[Enzyme, Reaction]):
    """ Relates an enzyme entry to a reaction which it catalyses.
    """
    # Attributes
    src: SourceDatabase  #: The source of the relation

    # Inverse
    inverse_main_relation = ReactionHasEnzyme

    # Database
    entrytype_name = 'Enzyme Catalyzes Reaction'
    _table_name = 'EnzymeCatalyzesReaction'
    symmetric = False
    source_entrytype = Enzyme
    target_entrytype = Reaction
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
        self,
        src: SourceDatabase = SourceDatabase.unknown,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src


    def _vis_str(self) -> str:
        return f'catalyses'

# Set inverse
ReactionHasEnzyme.has_inverse = EnzymeCatalyzesReaction



import re

re_comment = re.compile(r'{(.*)}')


def extract_molblocks_v2000(rxnblock: str) -> list[tuple[str, str]]:
    """ Gets the individual molblocks for each compound involved in the reaction. Returns tuples of the
    molblock and comment string.
    """
    molblocks: list[tuple[str, str]] = list()

    for mol_n, mol_str in enumerate(rxnblock.split('$MOL')):

        # Skip the first, $RXN block
        if mol_str.startswith('$RXN'):
            continue

        # Comment
        comment_match = re.search(re_comment, mol_str)
        if comment_match:
            comment_str = comment_match[0]
        else:
            comment_str = ''

        # Block
        lines = mol_str.splitlines()

        # First, read count line to get number of atoms
        countline_i = 0
        while len(lines[countline_i]) == 0 or (not lines[countline_i].strip()[0].isnumeric()):
            countline_i += 1
        countline = lines[countline_i]

        countline_items = filter(lambda x: x, countline.split(' '))
        n_atoms = int(countline[0:3])
        molblock = '\n'.join(lines[1:])

        molblocks.append((molblock, comment_str))

    # Finalise
    return molblocks






from __future__ import annotations

from typing import Optional

from chemrecon.chem.chemreaction import ChemReaction, chem_reaction_from_rxn
from chemrecon.chem.gml.gml_to_rdk import gml_to_rdkit_reaction
from chemrecon.core.id_types import A_GML_RULE, A_RXN
from chemrecon.schema import (ProceduralGeneratorError, ProceduralRelation, AAMRepr, AAM, )


class AAMConvert(ProceduralRelation[AAMRepr, AAM]):
    """ Standardized and canonical conversion of AAM representations (e.g. reaction SMILES, MolFile, …) into a
        consistent format, stored as a canonical reaction SMILES string.
    """
    # Attributes

    # Database
    entrytype_name = 'AAM Convert'
    _table_name = 'AAMConvert'
    symmetric = False
    source_entrytype = AAMRepr
    target_entrytype = AAM
    _attribute_columns = []
    _index = []

    def __init__(
        self,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)

    @classmethod
    def generate(
        cls,
        take_entry: AAMRepr
    ) -> list[tuple[ProceduralRelation[AAMRepr, AAM], AAM]]:

        r: Optional[ChemReaction] = None
        try:
            match take_entry.id_type:
                case A_RXN.enum_type:
                    r = chem_reaction_from_rxn(rxn = take_entry.source_id)
                case A_GML_RULE.enum_type:
                    rdk_reaction = gml_to_rdkit_reaction(gml = take_entry.source_id)
                    r = ChemReaction(
                        rdk_reaction = rdk_reaction
                    )
                case _:
                    # Log error and continue
                    raise ProceduralGeneratorError('Cannot create atom-to-atom map for reaction (unsupported format).')
        except Exception as e:
            raise ProceduralGeneratorError(f'Failed converting AAM representation to internal format.') from e
        if r is None:
            raise ProceduralGeneratorError(f'AAM conversion failed.')

        assert take_entry.recon_id is not None
        r_smarts = r.to_reaction_smiles()
        r_entry = AAM(
            reaction_smiles = r_smarts
        )

        # Return AAM
        return [
            (
                AAMConvert(),
                r_entry
            )
        ]



""" Functions for finding entries given an index.
"""
from typing import Optional

import chemrecon.connection as connection

from chemrecon import IdentifierType, Entry, IdentifierTypeCompound, IdentifierTypeReaction, IdentifierTypeEnzyme, \
    IdentifierTypeStructureRepresentation, IdentifierTypeAAM
from chemrecon.schema.entry_types.aam import AAM
from chemrecon.schema.entry_types.aam_repr import AAMRepr
from chemrecon.schema.entry_types.compound import Compound
from chemrecon.schema.entry_types.enzyme import Enzyme
from chemrecon.schema.entry_types.reaction import Reaction
from chemrecon.schema.entry_types.molstructure_repr import MolStructureRepr
from chemrecon.schema.entry_types.molstructure import MolStructure


def find_entry(id_type: IdentifierType, source_id: str) -> Optional[Entry]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    match id_type:
        case IdentifierTypeCompound():
            return find_compound_entry(id_type, source_id)
        case IdentifierTypeReaction():
            return find_reaction_entry(id_type, source_id)
        case IdentifierTypeEnzyme():
            return find_enzyme_entry(id_type, source_id)
        case IdentifierTypeStructureRepresentation():
            return find_structure_representation_entry(id_type, source_id)
        case IdentifierTypeAAM():
            return find_aam_representation_entry(id_type, source_id)
        case _:
            # Not implemented
            raise NotImplementedError()

def find_compound_entry(id_type: IdentifierTypeCompound, source_id: str) -> Optional[Compound]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    e = Compound(id_type = id_type.enum_type, source_id = id_type.std_identifier(source_id))
    return connection.handler.get_entry_by_index(e)

def find_reaction_entry(id_type: IdentifierTypeReaction, source_id: str) -> Optional[Reaction]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    e = Reaction(id_type = id_type.enum_type, source_id = id_type.std_identifier(source_id))
    return connection.handler.get_entry_by_index(e)

def find_enzyme_entry(id_type: IdentifierTypeEnzyme, source_id: str) -> Optional[Enzyme]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    e = Enzyme(id_type = id_type.enum_type, source_id = id_type.std_identifier(source_id))
    return connection.handler.get_entry_by_index(e)

def find_structure_representation_entry(id_type: IdentifierTypeStructureRepresentation, source_id: str) -> Optional[MolStructureRepr]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    e = MolStructureRepr(id_type = id_type.enum_type, source_id = id_type.std_identifier(source_id), implicit = False)
    return connection.handler.get_entry_by_index(e)

def find_aam_representation_entry(id_type: IdentifierTypeAAM, source_id: str) -> Optional[AAMRepr]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    e = AAMRepr(id_type = id_type.enum_type, source_id = id_type.std_identifier(source_id))
    return connection.handler.get_entry_by_index(e)

def find_structure_entry(smiles: str) -> Optional[MolStructure]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    e = MolStructure(smiles = smiles)
    return connection.handler.get_entry_by_index(e)

def find_aam_entry(reaction_smiles: str) -> Optional[AAM]:
    """ Look for an entry with the specified type and id in the connected Database.
        If not found, returns None.
    """
    e = AAM(reaction_smiles = reaction_smiles)
    return connection.handler.get_entry_by_index(e)



from __future__ import annotations

from typing import Optional

from rdkit import Chem as rdk

from chemrecon.chem.gml.gml_to_rdk import gml_to_rdkit_mol
from chemrecon.chem.mol import MolTemplate, MolInstance, feat_enum_map
from chemrecon.core.id_types import S_SMILES, S_INCHI, S_GML, S_MOLFILE
from chemrecon.chem.mol import Mol
from chemrecon.schema.entry_types.molstructure import MolStructure
from chemrecon.schema.entry_types.molstructure_repr import MolStructureRepr


def mol_from_smiles(smiles: str, provenance: Optional[str] = None, template: bool = False, safe = False) -> Mol:
    if template:
        mol = MolTemplate(
            rdk.MolFromSmiles(smiles),
            provenance = provenance
        )
    else:
        mol = MolInstance(
            rdk.MolFromSmiles(smiles),
            provenance = provenance
        )
    return mol


def mol_from_inchi(inchi: str, provenance: Optional[str] = None, template: bool = False, safe = False) -> Mol:
    if template:
        mol = MolTemplate(rdk.MolFromInchi(inchi), provenance = provenance)
    else:
        mol = MolInstance(rdk.MolFromInchi(inchi), provenance = provenance)
    return mol


def mol_from_molblock(molblock: str, provenance: Optional[str] = None, template: bool = False, safe = False) -> Mol:
    if template:
        mol = MolTemplate(
            rdk.MolFromMolBlock(molblock, sanitize = not safe, removeHs = not safe),
            provenance = provenance
        )
    else:
        mol = MolInstance(
            rdk.MolFromMolBlock(molblock, sanitize = not safe, removeHs = not safe),
            provenance = provenance
        )
    return mol


def mol_from_gml(gml: str, provenance: Optional[str] = None, template: bool = False) -> Mol:
    rdk_mol: rdk.Mol = gml_to_rdkit_mol(gml)
    mol = mol_from_rdk_mol(rdk_mol, provenance, template)
    return mol


def mol_from_rdk_mol(rdk_mol: rdk.Mol, provenance: Optional[str] = None, template: bool = False) -> Mol:
    if template:
        mol = MolTemplate(rdk_mol, provenance = provenance)
    else:
        mol = MolInstance(rdk_mol, provenance = provenance)
    return mol


def mol_from_structurerepresentation(
    entry: MolStructureRepr,
    provenance: Optional[str] = None,
    template: bool = False,
    safe: bool = False
):
    match entry.id_type:
        case S_SMILES.enum_type:
            return mol_from_smiles(entry.source_id, provenance, template = template, safe = safe)
        case S_INCHI.enum_type:
            return mol_from_inchi(entry.source_id, provenance, template = template, safe = safe)
        case S_MOLFILE.enum_type:
            return mol_from_molblock(entry.source_id, provenance, template = template, safe = safe)
        case S_GML.enum_type:
            return mol_from_gml(entry.source_id, provenance, template = template)
        case _:
            raise ValueError


def mol_from_struct_entry(entry: MolStructure):
    mol = Mol(
        rdk.MolFromSmiles(entry.smiles),
        set_features ={
            feat_enum_map[f] for f in entry.std_feats
        }
    )
    return mol



from typing import Optional

from chemrecon.schema import Relation, Reaction, SourceDatabase
from chemrecon.schema.db_object import col_src

class ReactionReference(Relation[Reaction, Reaction]):
    """ Inter- or intra-database reference between reactions.
    """
    # Attributes
    src: SourceDatabase  #: The source of the reference.

    # Database
    entrytype_name = 'Reaction Reference'
    _table_name = 'ReactionReference'
    symmetric = True
    source_entrytype = Reaction
    target_entrytype = Reaction
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
        self,
        src: SourceDatabase = SourceDatabase.unknown,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'ref ({self.src.name})'



from __future__ import annotations

from typing import Optional

from chemrecon.chem.elements import Element
from chemrecon.chem import elements as chem


class SumFormula:
    """ Encapsulates a molecular formula.
    """
    formula: dict[Element, int]
    charge: Optional[int]

    def __init__(self, formula: dict[Element, int], charge: Optional[int] = 0):
        """ Assumes 0 (neutral charge) rather than None (unknown). If charge is unknown, specify explicitly. """
        self.formula = formula
        self.charge = charge

    def __getitem__(self, element: Element):
        return self.formula[element]

    def __iter__(self):
        yield from self.formula.__iter__()

    def __or__(self, other: SumFormula):
        return self.formula | other.formula

    def get(self, element: Element, default: int):
        return self.formula.get(element, default)

    def elements(self) -> set[Element]:
        return set(self.formula.keys())

    def is_zero(self) -> bool:
        """ Returns true if all elements are 0, and charge is 0 or unknown.
        """
        if any(i != 0 for _, i in self.formula.items()):
            return False
        if self.charge is not None and self.charge != 0:
            return False
        return True

    def has_negative(self) -> bool:
        """ Returns true if the formula has any negative atom counts (e.g. when it represents a difference).
        """
        return any(v < 0 for v in self.formula.values())

    # Arithmetic operations
    def __add__(self, other: SumFormula):
        """ Add the elements and charges. If at least one has undefined charge, the result has undefined charge.
        """
        sumformula = {
            e: self.get(e, 0) + other.get(e, 0)
            for e in self.elements() | other.elements()
        }
        if self.charge is not None and other.charge is not None:
            charge: Optional[int] = self.charge + other.charge
        else:
            charge: Optional[int] = None

        return SumFormula(sumformula, charge)

    def __sub__(self, other: SumFormula):
        """ Computes the difference between sum formulae
        """
        sumformula = {
            e: self.get(e, 0) - other.get(e, 0)
            for e in self.elements() | other.elements()
        }
        if self.charge is not None and other.charge is not None:
            charge: Optional[int] = self.charge + other.charge
        else:
            charge: Optional[int] = None

        return SumFormula(sumformula, charge)

    def __eq__(self, other: SumFormula):
        # TODO how to handle equality when one has charge 0 and other has unknown charge?
        return self.formula == other.formula and self.charge == other.charge

    # Misc
    # ------------------------------------------------------------------------------------------------------------------
    def __str__(self):
        """ Prints output in Hill order
        """
        # https://en.wikipedia.org/wiki/Chemical_formula#Condensed_formula
        elist: list[tuple[Element, int]] = list()
        if chem.C in self.elements():
            # C, H first order
            elist.append((chem.C, self[chem.C]))
            if chem.H in self.elements():
                elist.append((chem.H, self[chem.H]))
            elist.extend([
                (e, i)
                for e, i in sorted(self.formula.items(), key = lambda pair: pair[0].symbol)
                if e not in {chem.C, chem.H}
            ])
        else:
            # Alphabetical order
            elist = [
                (e, i)
                for e, i in sorted(self.formula.items(), key = lambda pair: pair[0].symbol)
            ]

        chargestr: str = ''
        if self.charge is not None:
            if self.charge > 0:
                chargestr = f' {self.charge}+'
            elif self.charge < 0:
                chargestr = f' {abs(self.charge)}-'

        # Print as string
        return ''.join(f'{e.symbol}{i}' for e, i in elist) + f'{chargestr}'

def molformula_from_str(s: str) -> SumFormula:
    """ Convert a molecular formula string to an object representation.
    """
    # TODO
    raise NotImplementedError(s)



from typing import Optional

from chemrecon.schema.db_object import col_src
from chemrecon.schema import Reaction, AAM, ComposedRelation
from chemrecon.schema.enums import SourceDatabase
from chemrecon.schema.entry_types.aam_repr import AAMRepr
from chemrecon.schema.relation_types_procedural.aam_convert_relation import \
    AAMConvert
from chemrecon.schema.relation_types_source.reaction_has_aam_representation_relation import \
    ReactionHasAAMRepr


class ReactionHasAAM(ComposedRelation[Reaction, AAM, AAMRepr]):
    """ This relation gives the associated AAM of a reaction in a standardized format.
        This relation is composed of the `ReactionHasAAMRepr` relation and the `AAMConvert` relations.
        So if a reaction entry includes an RXN or GML representation of a map, this relation gives the same
        map in the standardized, ReactionSMILES-based format.
    """
    # Attributes
    src: SourceDatabase  #: The source database specifying this AAM.

    # Transitive
    rel_type_1 = ReactionHasAAMRepr
    rel_type_2 = AAMConvert
    intermediate_entrytype = AAMRepr

    # Database
    entrytype_name = 'Reaction has AAM'
    _table_name = 'ReactionHasAAM'
    symmetric = False
    source_entrytype = Reaction
    target_entrytype = AAM
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
            self,
            rel_1: ReactionHasAAMRepr,
            rel_2: AAMConvert,
            intermediate: AAMRepr,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        self.src = rel_1.src
        super().__init__(rel_1, rel_2, intermediate, recon_id_1, recon_id_2)

    def _vis_str(self) -> str:
        return f'{self.src.name}'



""" Handles the global connection state of ChemRecon.
"""
from __future__ import annotations

import os.path

import chemrecon.core.query_handler
import chemrecon.core.populate_query_handler
from chemrecon.database.connect import postgres_connect
from chemrecon.database.params import Params, local_docker_pub, local_docker_dev, chemrecon_pub, chemrecon_dev

# Handler, default to public production database
# ----------------------------------------------------------------------------------------------------------------------
handler: chemrecon.core.query_handler.QueryHandler  # QueryHandler or PopulateQueryHandler

def get_query_handler() -> chemrecon.core.query_handler.QueryHandler:
    """ Returns the current database query handler."""
    return handler

def connect(params: Params, can_write: bool):
    """ Sets the database handler for the ChemRecon library to a custom database connection.
        Also defines whether the library is allowed to write/cache results in the database.
    """
    global handler

    conn = postgres_connect(params)
    if can_write:
        handler = chemrecon.core.populate_query_handler.PopulateQueryHandler(conn)
    else:
        handler = chemrecon.core.query_handler.QueryHandler(conn)

    # Done
    print(f'Handler set: {params.connection_string()}')

def disconnect():
    global handler
    handler = None

def connect_public():
    """ Sets the ChemRecon library to use a database connection to the public
        ChemRecon database maintained by the developer.
    """
    connect(chemrecon_pub, can_write = False)

def connect_public_dev():
    """ For developer use.
    """
    connect(chemrecon_dev, can_write = True)

def connect_local_docker():
    """ Sets the ChemRecon library to use a database in a local Docker container. Refer to the documentation for
        details on how to run and administrate this database.
    """
    connect(local_docker_pub, can_write = False)

def connect_local_docker_dev():
    """ Sets the ChemRecon library to use a database in a local Docker container. Refer to the documentation for
        details on how to run and administrate this database.
    """
    connect(local_docker_dev, can_write = True)


# Local cache files location
# ----------------------------------------------------------------------------------------------------------------------
cache_dir: str = 'cache/'

def init_cache():
    """ If cache dir doesn't exis
    """
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

def flush_cache():
    """ Remove all cached results
    """
    raise NotImplementedError()

def set_cache_dir(new_cache_dir: str):
    global cache_dir
    cache_dir = new_cache_dir


# Initialise the cache
init_cache()

# External tools / dependencies
# ----------------------------------------------------------------------------------------------------------------------
# Reaction Decoder Tool, RDT
# rdt_jar_location: str = './chemrecon/redistribute/rdt/'
# rdt_jar_filename: str = 'rdt-2.4.1-jar-with-dependencies.jar'
# rdt_jar_path: str = './chemrecon/redistribute/rdt/rdt-2.4.1-jar-with-dependencies.jar'
# TODO by default in chemrecon/dependencies/rdt/...

# def set_rdt_jar_path(path: str):
#     # TODO
#     raise NotImplementedError()
#     pass



from collections import OrderedDict

from src.chemrecon import Direction
from chem.mol import MolTemplate, Mol
from chem.create_mol import mol_from_struct_entry
from chemrecon.entrygraph.explore import explore
from chemrecon.entrygraph.scoring import Scorer
from chemrecon.schema import (
    Compound, MolStructure
)
from chemrecon.schema.relation_types_procedural.compound_select_structure_proceduralrelation import \
    CompoundSelectStructure
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol


def get_structures_from_compound(
    entries: Compound | list[Compound],
    consider_first_entry_primary: bool = False,
) -> OrderedDict[MolTemplate, float]:

    # TODO update with new protocol syntax
    raise NotImplementedError()

    entry_set: set[Compound] = set()
    match entries:
        case Compound():
            entry_set = {entries}
        case list():
            entry_set = set(entries)
        case _:
            raise ValueError()

    # Construct entrygraph and run ranking  # TODO set first entry as primary if specified!
    try:
        eg = EG_Structure(
            initial_entries = entry_set
        )
    except ValueError as e:
        return OrderedDict()
    explore(entrygraph = eg, steps = 2)

    # Score
    scoring = scorer_default(eg)
    scoring_out: OrderedDict[Mol, float] = OrderedDict()
    for k, v in scoring.items():
        try:
            scoring_out[mol_from_struct_entry(k)] = v
        except AttributeError:
            continue

    return scoring_out


# Specification
EG_Structure = ExplorationProtocol(
    entry_types = {Compound, MolStructure},
    relation_types = {
        (CompoundSelectStructure, Direction.FORWARDS)
    },
    entry_types_initial = {Compound}
)

scorer_default = Scorer[MolStructure](
    score_entry_type = MolStructure
)


""" ChemRecon comes with a set of pre-defined exploration protocols for various purposes.
    These are located in the chemrecon.query.default_protocols module.
    We recommend looking in this file for inspiration on how to define custom protocols.
"""
from src.chemrecon import Direction
from chemrecon.entrygraph.filter import EntryFilterProcedure
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol
from chemrecon.schema.entry_types.compound import Compound
from chemrecon.schema.entry_types.molstructure_repr import MolStructureRepr
from chemrecon.schema.entry_types.molstructure import MolStructure
from chemrecon.schema.relation_types_composed.compound_has_molstructure_relation import CompoundHasMolStructure
from chemrecon.schema.relation_types_source.compound_reference_relation import CompoundReference
from chemrecon.schema.relation_types_source.molstructure_standardisation_relation import MolStructureStandardization

# TODO filters

# General
# ----------------------------------------------------------------------------------------------------------------------
#: The Compound-Structure protocol can be used to quickly gain an overview of the structural information
#: relating to a given compound.
#: The database compounds are traversed via the `CompoundReference` relation in order to expand the graph to include
#: other databases which contain the compound.
#: The `CompoundHasMolStructure` relation is then used to find the associated structure for each compound.
#: The `MolStructureStandardization` relation is used to standardize various properties of the structures, which can be
#: helpful in case the databases simply disagree on easy-to-standardize properties, such as charge or tautomerism.
protocol_compound_structure = ExplorationProtocol(
    relation_types = {
        (CompoundReference, Direction.SYMMETRIC),
        (CompoundHasMolStructure, Direction.FORWARDS),
        (MolStructureStandardization, Direction.FORWARDS),
    }
)

# Search graphs - for selection rather than manual inspection
# ----------------------------------------------------------------------------------------------------------------------
protocol_select_structure = ExplorationProtocol(
    relation_types = {
        (CompoundReference, Direction.SYMMETRIC),
        (CompoundHasMolStructure, Direction.FORWARDS),
        (MolStructureStandardization, Direction.FORWARDS),
    },
)

# TODO - FILTER - disallow implicit Structures!
def _select_structure_entry_filter(e: MolStructureRepr) -> bool:
    if e.implicit:
        return False
    else:
        return True

struct_repr_filter = EntryFilterProcedure[MolStructureRepr](
    filter_proc = _select_structure_entry_filter
)


# TODO include some ontology relations, such as the old_id and new_id relations




from typing import Optional

from chemrecon.schema import Entry
from chemrecon.schema.db_object import Column, col_recon_id

class AAM(Entry):
    """ Represents an atom-to-atom map.
    """

    # Attributes
    reaction_smiles: str  #: The canonical reaction SMILES string.

    # Database
    entrytype_name = 'AAM'
    _table_name = 'AAM'
    _columns = [
        col_recon_id,
        Column('reaction_smiles', str, index_hash = True),
    ]
    _index = [1]

    # Visualisation
    _draw_colour = '7AFF78'

    def __init__(
        self,
        reaction_smiles: str,
        recon_id: Optional[int] = None,
    ):
        super().__init__(recon_id)
        self.reaction_smiles = reaction_smiles

    def _vis_str(self) -> str:
        return self.reaction_smiles


from typing import Optional

from chemrecon.schema import Relation, Compound, SourceDatabase, MolStructureRepr
from chemrecon.schema.db_object import col_src


class CompoundHasStructureRepresentation(Relation[Compound, MolStructureRepr]):
    """ Relates a compound entry to the structure representation (e.g. SMILES, InChI) given by that database.
    """
    # Attributes
    src: SourceDatabase  #: The source database containing this representation.

    # Database
    _table_name = 'CompoundHasStructureRepresentation'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = MolStructureRepr
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    def __init__(
        self,
        src: SourceDatabase = SourceDatabase.unknown,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.src = src

    def _vis_str(self) -> str:
        return f'{self.src.name}'



from typing import Any


class EncodableList[T: Any](list[T]):
    """ Simple wrapper for built-in list class to allow for encoding for memoization purposes.
    """
    def encode(self, *args, **kwargs):
        return str([x.encode() for x in self.__iter__()]).encode()

    def __hash__(self):
        return self.encode().__hash__()



import os

def set_cwd():
    while os.getcwd().split('/')[-1] not in {
        'chemrecon',
        'ChemRecon',
        'src',
        'chemrecon_populator'
    }:
        os.chdir('..')

    print(f'Changed directory: {os.getcwd()}')



from typing import Optional

from chemrecon.chem.mol import feats, Mol
from chemrecon.chem.create_mol import mol_from_smiles
from chemrecon.schema import Column, FeatureEnum, ProceduralGeneratorError, ProceduralRelation, MolStructure


class MolStructureStandardization(ProceduralRelation[MolStructure, MolStructure]):
    """ Standardization of molecular structures according to a particular 'feature' (Fragment, Isotope, Charge,
        Tautomerism, Stereochemical).
    """

    # Attributes
    feat: FeatureEnum  #: The feature w.r.t. which the structure is standardized.

    # Database
    entrytype_name = 'MolStructure standardisation'
    _table_name = 'MolStructureStandardisation'
    ignore_generation_limit = True
    symmetric = False
    source_entrytype = MolStructure
    target_entrytype = MolStructure
    _attribute_columns = [
        Column('feat', FeatureEnum)
    ]
    _index = [0]

    def __init__(
        self,
        feat: FeatureEnum,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.feat = feat

    @classmethod
    def generate(
        cls, take_entry: MolStructure
    ) -> list[tuple[ProceduralRelation[MolStructure, MolStructure], MolStructure]]:

        # Create mol object
        m: Mol
        try:
            m = mol_from_smiles(take_entry.smiles)
        except ValueError as e:
            # Mol creation fails
            raise ProceduralGeneratorError('Could not create mol for std.') from e

        feat_entries: list[tuple[FeatureEnum, MolStructure]] = list()
        for f in feats:
            if f not in m.features:
                try:
                    m_: Mol = Mol(f.standardise(m))
                except Exception as e:
                    # Failed standardisation
                    continue

                feat_entries.append((f.feature_enum, m_.to_database_struct()))

        return [
            (
                MolStructureStandardization(feat = f),
                e
            )
            for f, e in feat_entries
        ]

    # Visualisation
    def _vis_str(self) -> str:
        return f'std: {self.feat.name}'
